import re
import subprocess

from collections import namedtuple
import six

from . import utils

try:
    from subprocess import DEVNULL
except ImportError:
    import os
    DEVNULL = open(os.devnull, 'wb')


def get_recommended_repo(path):
    """Given vcs directory ``path``, returns the url from which the repo
    can be cloned.

    For git, an 'upstream' remote will be preferred over 'origin'.
    For bzr, the :parent: branch will be preferred.
    For hg, the 'default' alias will be preferred.

    Returns None if the directory is not a repo, or a remote url can not
    be determined.

    :param path: A :class:`path.path` to a directory
    :return: A url string, or None

    """

    Command = namedtuple("Command", "args parse")
    cmds = [
        Command(['git', 'remote', '-v'], _parse_git),
        Command(['bzr', 'info'], _parse_bzr),
        Command(['hg', 'paths'], _parse_hg),
    ]

    if not path.exists():
        return None

    with utils.cd(str(path)):
        for cmd in cmds:
            try:
                output = subprocess.check_output(cmd.args, stderr=DEVNULL)
                if six.PY3:
                    output = str(output)
                if output:
                    repo = cmd.parse(output)
                    if repo:
                        return repo
            except (subprocess.CalledProcessError, OSError):
                continue


def _parse_git(txt):
    pat = re.compile(
        r'(?P<name>\S+)\s+(?P<url>\S+)\s+\((?P<type>[^\)]+)\)')
    urls = {}
    for line in txt.split('\n'):
        match = pat.search(line)
        if match:
            d = match.groupdict()
            if d['name'] == 'upstream' and d['type'] == 'fetch':
                return d['url'].strip()
            elif d['type'] == 'fetch':
                urls[d['name']] = d['url'].strip()

    if 'origin' in urls:
        return urls['origin']

    for url in urls.values():
        return url


def _parse_bzr(txt):
    branch_types = ['parent', 'push', 'submit']
    pat = re.compile(
        r'(?P<branch_type>({})) branch: (?P<url>.*)'.format(
            '|'.join(branch_types)))
    matches = {}
    for line in txt.split('\n'):
        match = pat.search(line)
        if match:
            d = match.groupdict()
            matches[d['branch_type']] = d['url'].strip()
    if not matches:
        return
    for typ in branch_types:
        url = matches.get(typ)
        if url:
            return url


def _parse_hg(txt):
    pat = re.compile(r'(?P<name>[^\s]+) = (?P<url>.*)')
    urls = []
    for line in txt.split('\n'):
        match = pat.search(line)
        if match:
            d = match.groupdict()
            if d['name'] == 'default':
                return d['url'].strip()
            else:
                urls.append(d['url'].strip())
    return urls[0] if urls else None



#!/usr/bin/python

from charmhelpers.core.services.base import ServiceManager
from charmhelpers.core.services import helpers

import actions


def manage():
    manager = ServiceManager([
        {
            'service': '$metadata.package',
            'ports': [],  # ports to after start
            'provided_data': [
                # context managers for provided relations
                # e.g.: helpers.HttpRelation()
            ],
            'required_data': [
                # data (contexts) required to start the service
                # e.g.: helpers.RequiredConfig('domain', 'auth_key'),
                #       helpers.MysqlRelation(),
            ],
            'data_ready': [
                helpers.render_template(
                    source='upstart.conf',
                    target='/etc/init/$metadata.package'),
                actions.log_start,
            ],
        },
    ])
    manager.manage()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import blessings
import curses
import json
import logging
import os
import requests
import sys
import uuid
import yaml
import string
import traceback

import charmtools.build.tactics

from path import Path as path
from collections import OrderedDict
from charmtools import (utils, repofinder, proof)
from charmtools.build import inspector
from charmtools.build.errors import BuildError
from charmtools.build.config import BuildConfig, DEFAULT_IGNORES
from charmtools.build.tactics import Tactic, WheelhouseTactic
from charmtools.build.fetchers import (
    InterfaceFetcher,
    LayerFetcher,
    get_fetcher,
    FetchError,
    RepoFetcher,
)
from charmtools.version import cached_charm_tools_version, format_version
from zipfile import ZipFile, ZIP_DEFLATED

log = logging.getLogger("build")


class Configable(object):
    CONFIG_FILE = None
    OLD_CONFIG_FILE = None

    def __init__(self):
        self._config = BuildConfig()
        self.config_file = self.CONFIG_FILE
        self.old_config_file = self.OLD_CONFIG_FILE

    @property
    def config(self):
        if self._config.configured:
            return self._config
        if self.config_file and self.config_file.exists():
            self._config.configure(self.config_file)
        elif self.old_config_file and self.old_config_file.exists():
            self._config.configure(self.old_config_file)
        return self._config

    @property
    def configured(self):
        return bool(self.config is not None and self.config.configured)


class Fetched(Configable):
    def __init__(self, url, target_repo, name=None, lock=None,
                 use_branches=False):
        super(Fetched, self).__init__()
        self.url = url
        self.target_repo = target_repo / self.NAMESPACE
        self.directory = None
        self._name = name
        self.fetched = False
        self.fetched_url = None
        self.vcs = None
        self.revision = None
        self.branch = None
        self.lock = lock
        self.use_branches = use_branches

    @property
    def name(self):
        if self._name:
            return self._name
        if self.url.startswith(self.NAMESPACE + ':'):
            return self.url[len(self.NAMESPACE)+1:]
        return self.url.rsplit('/', 1)[-1]

    def __repr__(self):
        return "<{} {}:{}>".format(self.__class__.__name__,
                                   self.url, self.directory)

    def __div__(self, other):
        return self.directory / other

    def fetch(self):
        try:
            # In order to lock the fetcher we need to adjust the self.url
            # to get the right thing.  Curiously, self.url is actually
            # "layer:something" here, and so we can match on that.
            if self.lock:
                url = make_url_from_lock_for_layer(
                    self.lock, self.use_branches)
            else:
                url = self.url
            fetcher = get_fetcher(url)
        except FetchError:
            # We might be passing a local dir path directly
            # which fetchers don't currently  support
            self.directory = path(self.url)
            self.revision = RepoFetcher(self.url).get_revision(self.url)
        else:
            if hasattr(fetcher, "path") and fetcher.path.exists():
                self.directory = path(fetcher.path)
            else:
                if not self.target_repo.exists():
                    self.target_repo.makedirs_p()
                self.directory = path(fetcher.fetch(self.target_repo))
                self.fetched = True
                self.fetched_url = getattr(fetcher, "fetched_url", None)
                self.vcs = getattr(fetcher, "vcs", None)
            self.revision = fetcher.get_revision(self.directory)
            self.branch = fetcher.get_branch_for_revision(self.directory,
                                                          self.revision)

        if not self.directory.exists():
            raise BuildError(
                "Unable to locate {}. "
                "Do you need to set {}?".format(
                    self.url, self.ENVIRON))

        if self.config_file:
            self.config_file = self.directory / self.config_file
        if self.old_config_file:
            self.old_config_file = self.directory / self.old_config_file
        self._name = self.config.name
        return self


class Interface(Fetched):
    CONFIG_FILE = "interface.yaml"
    NAMESPACE = "interface"
    ENVIRON = "CHARM_INTERFACES_DIR"
    OLD_ENVIRON = "INTERFACE_PATH"


class Layer(Fetched):
    CONFIG_FILE = "layer.yaml"
    OLD_CONFIG_FILE = "composer.yaml"
    NAMESPACE = "layer"
    ENVIRON = "CHARM_LAYERS_DIR"
    OLD_ENVIRON = "LAYER_PATH"


class Builder(object):
    """
    Handle the processing of overrides, implements the policy of BuildConfig
    """
    PHASES = ['lint', 'read', 'call', 'sign', 'build']
    HOOK_TEMPLATE_FILE = path('hooks/hook.template')
    DEFAULT_SERIES = 'trusty'
    METRICS_URL = 'https://www.google-analytics.com/collect'
    METRICS_ID = 'UA-96529618-2'

    def __init__(self):
        self.config = BuildConfig()
        self.force = False
        self._name = None
        self._charm = None
        self._top_layer = None
        self.hide_metrics = os.environ.get('CHARM_HIDE_METRICS', False)
        self.wheelhouse_overrides = None
        self.wheelhouse_per_layer = False
        self._warned_home = False
        self.lock_items = []
        self.with_locks = {}

    @property
    def top_layer(self):
        if not self._top_layer:
            # NOTE: no lock for branches for lop layer.
            self._top_layer = Layer(self.charm, self.cache_dir).fetch()

        return self._top_layer

    @property
    def charm(self):
        return self._charm

    @charm.setter
    def charm(self, value):
        self._charm = path(value)

    @property
    def charm_metadata(self):
        if not hasattr(self, '_charm_metadata'):
            md = path(self.charm) / "metadata.yaml"
            try:
                setattr(
                    self, '_charm_metadata',
                    yaml.safe_load(md.open()) if md.exists() else None)
            except yaml.YAMLError as e:
                log.debug(e)
                raise BuildError("Failed to process {0}. "
                                 "Ensure the YAML is valid".format(md))

        return self._charm_metadata

    @property
    def name(self):
        if self._name:
            return self._name

        if self.charm is None:
            return None

        # optionally extract name from the top layer
        self._name = str(path(self.charm).abspath().name)

        # however if the current layer has a metadata.yaml we can
        # use its name
        if self.charm_metadata:
            name = self.charm_metadata.get("name")
            if name:
                self._name = name
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def source_dir(self):
        return self.top_layer.directory

    @property
    def target_dir(self):
        return self.build_dir / self.name

    @property
    def manifest(self):
        return self.target_dir / '.build.manifest'

    @property
    def lock_file(self):
        return self.source_dir / 'build.lock'

    def check_series(self):
        """Make sure this is a either a multi-series charm, or we have a
        build series defined. If not, fall back to a default series.

        """
        if self.charm_metadata and self.charm_metadata.get('series'):
            return
        elif self.series:
            log.warn('DEPRECATED: use of --series flag; '
                     'specify series in metadata.yaml instead')
            return
        else:
            log.warn('DEPRECATED: implicit series; '
                     'specify series in metadata.yaml instead')

    def status(self):
        result = {}
        result.update(vars(self))
        for e in ["CHARM_LAYERS_DIR",
                  "CHARM_INTERFACES_DIR",
                  "CHARM_BUILD_DIR",
                  "JUJU_REPOSITORY",
                  "LAYER_PATH",
                  "INTERFACE_PATH"]:
            result[e] = os.environ.get(e)
        return result

    def find_or_create_target(self):
        log.info("Destination charm directory: {}".format(self.target_dir))
        self.target_dir.makedirs_p()

    @property
    def layers(self):
        layers = []
        for layer in (self._layers, self._interfaces):
            for i in layer:
                item = {
                    'url': i.url,
                    'rev': i.revision,
                }
                for opt in ('branch', 'fetched_url'):
                    attr = getattr(i, opt, None)
                    if attr:
                        item[opt] = attr
                layers.append(item)
        return layers

    def fetch(self):
        self.target_dir.makedirs_p()
        if not self.top_layer.configured:
            log.warn("The top level layer expects a "
                     "valid layer.yaml file")
        # Manually create a layer object for the output
        self.target = Layer(self.name, self.build_dir,
                            lock=self.lock_for(self.name),
                            use_branches=getattr(self,
                                                 'use_lock_file_branches',
                                                 False))
        self.target.directory = self.target_dir
        self.post_metrics('charm', self.name, False)
        return self.fetch_deps(self.top_layer)

    def fetch_deps(self, layer):
        self.cache_dir.makedirs_p()
        results = {"layers": [], "interfaces": []}
        self.fetch_dep(layer, results)
        # results should now be a bottom up list
        # of deps. Using the in order results traversal
        # we can build out our plan for each file in the
        # output layer
        results["layers"].append(layer)
        self._layers = results["layers"]
        self._interfaces = results["interfaces"]
        return results

    def fetch_dep(self, layer, results):
        # Recursively fetch and scan layers
        # This returns a plan for each file in the result
        baselayers = layer.config.get('includes', [])
        if not baselayers:
            # no deps, this is possible for any base
            # but questionable for the target
            return

        if isinstance(baselayers, str):
            baselayers = [baselayers]

        for base in baselayers:
            # The order of these commands is important. We only want to
            # fetch something if we haven't already fetched it.
            if base.startswith("interface:"):
                iface = Interface(base, self.cache_dir,
                                  lock=self.lock_for(base),
                                  use_branches=getattr(
                                      self, 'use_lock_file_branches', False))
                if iface.name in [i.name for i in results['interfaces']]:
                    continue
                results["interfaces"].append(iface.fetch())
                self.post_metrics('interface', iface.name, iface.fetched)
            else:
                base_layer = Layer(base, self.cache_dir,
                                   lock=self.lock_for(base),
                                   use_branches=getattr(
                                       self, 'use_lock_file_branches', False))
                if base_layer.name in [i.name for i in results['layers']]:
                    continue
                base_layer.fetch()
                self.fetch_dep(base_layer, results)
                results["layers"].append(base_layer)
                self.post_metrics('layer', base_layer.name, base_layer.fetched)

    def lock_for(self, base):
        """Return a lock description for an item 'base' if it exists."""
        return self.with_locks.get(base, {})

    def build_tactics(self, entry, layer, next_config, current_config,
                      output_files):
        relname = entry.relpath(layer.directory)
        existing_tactic = output_files.get(relname)
        tactic = Tactic.get(
            entry,
            self.target,
            layer,
            next_config,
            current_config,
            existing_tactic,
        )
        output_files[relname] = tactic

    def plan_layers(self, layers, output_files):
        current_config = BuildConfig()
        next_config = current_config.add_config(layers["layers"][0].config)

        layers["layers"][-1].url = self.name

        for i, layer in enumerate(layers["layers"]):
            log.info("Processing layer: %s%s", layer.url,
                     "" if layer.directory.startswith(self.cache_dir)
                     else " (from %s)" % layer.directory.relpath())
            current_config = current_config.add_config(layer.config)
            if i + 1 < len(layers["layers"]):
                next_layer = layers["layers"][i + 1]
                next_config = next_config.add_config(next_layer.config)
            else:
                # Add an empty level to the configs to represent that there
                # is no layer after the current one.  This is important for
                # the IgnoreTactic, which needs to look ahead so that it can
                # handle ignoring entire directories.
                next_config = next_config.add_config({})
            list(e for e in utils.walk(layer.directory,
                                       self.build_tactics,
                                       layer=layer,
                                       next_config=next_config,
                                       current_config=current_config,
                                       output_files=output_files))
        # now we do update the wheelhouse.txt output file with the lock file if
        # necessary.
        if not getattr(self, 'ignore_lock_file', False):
            lines = self.generate_python_modules_from_lock_file()
            # override any existing lines with the python modules from the lock
            # file.
            existing_tactic = output_files.get('wheelhouse.txt')
            lock_layer = Layer('lockfile-wheelhouse',
                               layers["layers"][-1].target_repo.dirname())
            lock_layer.directory = layers["layers"][-1].directory
            wh_tactic = WheelhouseTactic(
                "",
                self.target,
                lock_layer,
                next_config,
            )
            wh_tactic.lines = lines
            wh_tactic.purge_wheels = True
            if existing_tactic is not None:
                wh_tactic.combine(existing_tactic)
            output_files["wheelhouse.txt"] = wh_tactic

        if self.wheelhouse_overrides:
            existing_tactic = output_files.get('wheelhouse.txt')
            wh_over_layer = Layer('--wheelhouse-overrides',
                                  layers["layers"][-1].target_repo.dirname())
            wh_over_layer.directory = layers["layers"][-1].directory
            output_files['wheelhouse.txt'] = WheelhouseTactic(
                self.wheelhouse_overrides,
                self.target,
                wh_over_layer,
                next_config,
            )
            output_files['wheelhouse.txt'].purge_wheels = True
            if existing_tactic is not None:
                output_files['wheelhouse.txt'].combine(existing_tactic)
        plan = [t for t in output_files.values() if t]
        return plan

    def plan_hooks(self, layers, output_files, plan):
        """
        Add a Tactic to the plan to handle rendering all standard hooks from
        the template, if not explicitly overridden.

        :param layers: Data structure containing all of the layers composing
            this charm, along with some overall metadata.
        :type layers: dict
        :param output_files: Mapping of file Paths to Tactics that should
            process those files.
        :type output_files: dict
        :param plan: List of all Tactics that need to be invoked.
        :type plan: list
        """
        if self.HOOK_TEMPLATE_FILE not in output_files:
            raise BuildError('At least one layer must provide %s',
                             self.HOOK_TEMPLATE_FILE)
        template_file = self.target.directory / self.HOOK_TEMPLATE_FILE
        target_config = layers["layers"][-1].config
        source_layer = output_files[self.HOOK_TEMPLATE_FILE].layer
        plan.append(
            charmtools.build.tactics.StandardHooksBind(
                'hook', source_layer.url, self.target,
                target_config, output_files, template_file))

    def plan_interfaces(self, layers, output_files, plan):
        """
        Add Tactics to the plan for each relation endpoint to render hooks
        for that relation endpoint from the template, as well as a tactic
        to pull in the interface layer's code (if there is an interface layer).

        :param layers: Data structure containing all of the layers composing
            this charm, along with some overall metadata.
        :type layers: dict
        :param output_files: Mapping of file Paths to Tactics that should
            process those files.
        :type output_files: dict
        :param plan: List of all Tactics that need to be invoked.
        :type plan: list
        """
        # Interface includes don't directly map to output files
        # as they are computed in combination with the metadata.yaml
        metadata_tactic = [tactic for tactic in plan if isinstance(
                           tactic, charmtools.build.tactics.MetadataYAML)]
        if not metadata_tactic:
            raise BuildError('At least one layer must provide '
                             'metadata.yaml')
        meta = metadata_tactic[0].process()
        if not meta and layers.get('interfaces'):
            raise BuildError(
                'Includes interfaces but no metadata.yaml to bind them')
        elif self.HOOK_TEMPLATE_FILE not in output_files:
            raise BuildError('At least one layer must provide %s',
                             self.HOOK_TEMPLATE_FILE)
        elif not meta:
            log.warn('Empty metadata.yaml')

        template_file = path(self.target.directory) / self.HOOK_TEMPLATE_FILE
        target_config = layers["layers"][-1].config
        specs = []
        used_interfaces = set()
        for role in ("provides", "requires", "peers"):
            for k, v in meta.get(role, {}).items():
                # ex: ["provides", "db", "mysql"]
                specs.append([role, k, v["interface"]])
                used_interfaces.add(v["interface"])

        for iface in layers.get("interfaces", []):
            if iface.name not in used_interfaces:
                # we shouldn't include something the charm doesn't use
                log.warn("layer.yaml includes {} which isn't "
                         "used in metadata.yaml".format(
                             iface.name))
                continue
            for role, relation_name, interface_name in specs:
                if interface_name != iface.name:
                    continue

                log.info("Processing interface: %s%s", interface_name,
                         "" if iface.directory.startswith(self.cache_dir)
                         else " (from %s)" % iface.directory.relpath())
                # COPY phase
                plan.append(
                    charmtools.build.tactics.InterfaceCopy(
                        iface, relation_name, role,
                        self.target, target_config)
                )

        # Link Phase
        # owner = metadata_tactic[0].layer.url
        owner = output_files[self.HOOK_TEMPLATE_FILE].layer.url
        for role, relation_name, interface_name in specs:
            plan.append(charmtools.build.tactics.InterfaceBind(
                relation_name, owner, self.target,
                target_config, output_files, template_file))

    def plan_storage(self, layers, output_files, plan):
        """
        Add Tactics to the plan for each storage endpoint to render hooks
        for that storage endpoint from the template.

        :param layers: Data structure containing all of the layers composing
            this charm, along with some overall metadata.
        :type layers: dict
        :param output_files: Mapping of file Paths to Tactics that should
            process those files.
        :type output_files: dict
        :param plan: List of all Tactics that need to be invoked.
        :type plan: list
        """
        # Storage hooks don't directly map to output files
        # as they are computed in combination with the metadata.yaml
        metadata_tactic = [tactic for tactic in plan if isinstance(
                           tactic, charmtools.build.tactics.MetadataYAML)]
        if not metadata_tactic:
            raise BuildError('At least one layer must provide metadata.yaml')
        meta_tac = metadata_tactic[0]
        meta_tac.process()
        if not meta_tac.storage:
            return
        if self.HOOK_TEMPLATE_FILE not in output_files:
            raise BuildError('At least one layer must provide %s',
                             self.HOOK_TEMPLATE_FILE)

        template_file = path(self.target.directory) / self.HOOK_TEMPLATE_FILE
        target_config = layers["layers"][-1].config
        for name, owner in meta_tac.storage.items():
            plan.append(
                charmtools.build.tactics.StorageBind(
                    name, owner, self.target,
                    target_config, output_files, template_file))

    def plan_version(self, layers, output_files, plan):
        plan.append(
            charmtools.build.tactics.VersionTactic(
                self.charm,
                self.target,
                self.top_layer, None,
            ))

    def formulate_plan(self, layers):
        """Build out a plan for each file in the various
        layers, taking into account config at each layer"""
        output_files = OrderedDict()
        self.plan = self.plan_layers(layers, output_files)
        self.plan_hooks(layers, output_files, self.plan)
        self.plan_interfaces(layers, output_files, self.plan)
        self.plan_storage(layers, output_files, self.plan)
        self.plan_version(layers, output_files, self.plan)
        return self.plan

    def post_metrics(self, kind, layer_name, fetched):
        if self.hide_metrics:
            return
        conf_file = path('~/.config/charm-build.conf').expanduser()
        try:
            conf = yaml.safe_load(conf_file.text()) or {}
        except (FileNotFoundError, yaml.error.YAMLError):
            conf = {}
        if not conf.get('cid'):
            conf_file.parent.makedirs_p()
            conf['cid'] = str(uuid.uuid4())
            conf_file.write_text(yaml.safe_dump(conf))
        cid = conf['cid']
        try:
            requests.post(self.METRICS_URL, timeout=10, data={
                'tid': self.METRICS_ID,
                'v': 1,
                'aip': 1,
                't': 'event',
                'ds': 'app',
                'cid': cid,
                'av': format_version(cached_charm_tools_version(), 'long'),
                'an': "charm-build",
                'ec': kind,
                'ea': 'fetch' if fetched else 'local',
                'el': layer_name,
                'cd1': self.series,
            })
        except requests.exceptions.RequestException:
            pass

    def exec_plan(self, plan=None, layers=None):
        signatures = {}
        cont = True
        for phase in self.PHASES:
            for tactic in plan:
                if phase == "lint":
                    cont &= tactic.lint()
                    if cont is False and self.force is not True:
                        # no message, reason will already have been logged
                        raise BuildError()
                elif phase == "read":
                    # We use a read (into memory phase to make layer comps
                    # simpler)
                    tactic.read()
                elif phase == "call":
                    tactic()
                    if hasattr(tactic, "lock_info"):
                        self.lock_items.extend(tactic.lock_info)
                elif phase == "sign":
                    sig = tactic.sign()
                    if sig:
                        signatures.update(sig)
        new_repo = not self.manifest.exists()
        if new_repo:
            added, changed, removed = set(), set(), set()
        else:
            added, changed, _ = utils.delta_signatures(self.manifest)
            removed = self.clean_removed(signatures)
        # write out the sigs
        if "sign" in self.PHASES:
            self.write_signatures(signatures, layers)
        if getattr(self, 'write_lock_file', False):
            self.write_the_lock_file()
        if self.report:
            self.write_report(new_repo, added, changed, removed)

    def write_signatures(self, signatures, layers):
        signatures['.build.manifest'] = ["build", 'dynamic', 'unchecked']
        self.manifest.write_text(json.dumps(dict(
            signatures=signatures,
            layers=layers,
        ), indent=2, sort_keys=True))

    def write_the_lock_file(self):
        """Using the info in self.layers, write a lock file.

        The lock file can be used to ensure that the same versions (or
        branches) are used to recreate the charm.
        """
        locks = []
        # look at the layers first for locks
        for layer in self.layers:
            branch = layer.get('branch', None)
            if branch is None or branch == 'refs/heads/master':
                tag = layer.get('rev', None)
            else:
                tag = branch
            lock_item = {
                'type': 'layer',
                'item': layer['url'],
                'url': layer.get('fetched_url', None),
                'vcs': layer.get('vcs', None),
                'branch': branch,
                'commit': tag
            }
            locks.append(lock_item)
        # now iterate through the other lock items and add them in
        locks.extend(self.lock_items)
        self.lock_file.write_text(json.dumps({'locks': locks}, indent=2))

    def maybe_read_lock_file(self):
        """Read the lock file if it exists.

        The lock file is a list of elements (either layer, interface or python
        module).  This is then used in the fetcher system to fix the versions
        that are used when rebuilding the charm.
        """
        self.with_locks = {}
        try:
            with open(self.lock_file) as f:
                with_locks = json.load(f)
        except FileNotFoundError:
            log.info("The lockfile %s was not found; building using latest "
                     "versions.", self.lock_file)
            return
        except Exception as err:
            log.error("Problem decoding the lock file: %s\n%s",
                      self.lock_file, str(err))
            raise
        # re-format the lock-file as a dictionary so that the items can be
        # looked up.  A layer is 'layer:<name>', an interface is
        # 'interface:<name>', and a python moudle is 'python_module:<name>'
        for item in with_locks['locks']:
            if item['type'] == 'python_module':
                self.with_locks["python_module:{}".format(item['package'])] = \
                    item
            elif item['type'] == 'layer':
                self.with_locks[item['item']] = item
            else:
                log.warning("Not sure how to deal with lock item '%s'?", item)
        log.info("Using lockfile %s for build.", self.lock_file)

    def generate_python_modules_from_lock_file(self):
        """Using self.with_locks[], generate a list of python module lines.

        This takes the python modules recorded in the lock file, as read into
        self.with_locks[] and generate a list of python module line in the
        style of a requiremens.txt file suitable for pip.  These are absolute
        version locks (or github commit/branch locks).

        :returns: list of python module lines
        :rtype: List[str]
        """
        lines = []
        for data in self.with_locks.values():
            if data['type'] != "python_module":
                continue
            vcs = data['vcs']
            if vcs == 'git':
                if self.use_lock_file_branches:
                    branch = data.get('branch', '')
                    if branch.startswith("refs/heads/"):
                        branch = branch[len("refs/heads/"):]
                    line = "{}@{}#egg={}".format(
                        data['url'], branch, data['package'])
                else:
                    line = "{}@{}#egg={}".format(
                        data['url'], data['version'], data['package'])
            elif vcs is None:
                line = "{}=={}".format(data['package'], data['version'])
            else:
                log.error("Unknown vcs type %s - aborting.", vcs)
                sys.exit(1)
            lines.append(line)
        return lines

    def generate(self):
        layers = self.fetch()
        self.formulate_plan(layers)
        self.exec_plan(self.plan, self.layers)

    def validate(self):
        if not (self.name and str(self.name)[0] in string.ascii_lowercase):
            raise BuildError('Charm name must start with a lower-case letter')

        for cls in (InterfaceFetcher, LayerFetcher):
            if cls.OLD_ENVIRON in os.environ and cls.ENVIRON not in os.environ:
                log.warning('DEPRECATED: {} environment variable; '
                            'please use {} instead'.format(cls.OLD_ENVIRON,
                                                           cls.ENVIRON))

        self._validate_charm_repo()

        if not self.manifest.exists():
            return [], [], []
        a, c, d = utils.delta_signatures(self.manifest, ignore=DEFAULT_IGNORES)

        for f in a:
            log.warn("Conflict: File in destination directory "
                     "was added after charm build: %s", f)
        for f in c:
            log.warn("Conflict: File in destination directory "
                     "was modified after charm build: %s", f)
        for f in d:
            log.warn("Conflict: File in destination directory "
                     "was deleted after charm build: %s", f)
        if a or c or d:
            if self.force is True:
                log.info(
                    "Continuing with known changes to target layer. "
                    "Changes will be overwritten")
            else:
                raise BuildError(
                    "Unable to continue due to unexpected modifications "
                    "(try --force)")

        return a, c, d

    def _validate_charm_repo(self):
        if 'repo' not in self.top_layer.config:
            msg = 'Please add a `repo` key to your {}'.format(
                self.top_layer.config_file.name)
            recommended_repo = repofinder.get_recommended_repo(self.charm)
            if recommended_repo:
                msg += ', e.g. repo: {}'.format(recommended_repo)
            else:
                msg += ', with a url from which your layer can be cloned.'
            log.warn(msg)

    def create_charm_file(self):
        dst_name = '{}.charm'.format(self.name)
        with ZipFile(dst_name, 'w', ZIP_DEFLATED) as zip:
            for src_path in self.target_dir.walkfiles():
                zip.write(src_path, src_path.relpath(self.target_dir))

    def __call__(self):
        log.debug(json.dumps(
            self.status(), indent=2, sort_keys=True, default=str))
        self.validate()
        self.find_or_create_target()
        self.generate()
        if self.charm_file:
            self.create_charm_file()
        self.cleanup()

    def inspect(self):
        self.charm = path(self.charm).abspath()
        self._check_path(self.charm)
        inspector.inspect(self.charm,
                          force_styling=self.force_color,
                          annotate=self.annotate)

    def normalize_build_dir(self):
        charm_build_dir = os.environ.get('CHARM_BUILD_DIR')
        juju_repo_dir = os.environ.get('JUJU_REPOSITORY')
        series = self.series or 'builds'
        if not self.build_dir:
            if self.output_dir:
                self.build_dir = self.output_dir
            elif charm_build_dir:
                self.build_dir = path(charm_build_dir)
            elif juju_repo_dir:
                self.build_dir = path(juju_repo_dir) / series
            else:
                log.warn('Build dir not specified via command-line or '
                         'environment; defaulting to /tmp/charm-builds')
                self.build_dir = path('/tmp/charm-builds')
        self.build_dir = self.build_dir.abspath()
        charm_dir = path(self.charm).abspath() + os.path.sep
        if os.path.commonprefix([self.build_dir, charm_dir]) == charm_dir:
            raise BuildError('Build directory nested under source directory. '
                             'This will cause recursive nesting of build '
                             'artifacts and can fill up your disk. Please '
                             'specify a different build directory with '
                             '--build-dir or $CHARM_BUILD_DIR')

    def normalize_cache_dir(self):
        charm_cache_dir = os.environ.get('CHARM_CACHE_DIR')
        if not self.cache_dir:
            if charm_cache_dir:
                self.cache_dir = path(charm_cache_dir)
            else:
                self.cache_dir = path('~/.cache/charm').expanduser()
        self.cache_dir = self.cache_dir.abspath() / str(os.getpid())
        if self.cache_dir.startswith(path(self.charm).abspath()):
            raise BuildError('Cache directory nested under source directory. '
                             'This will cause recursive nesting of build '
                             'artifacts and can fill up your disk. Please '
                             'specify a different build directory with '
                             '--cache-dir or $CHARM_CACHE_DIR')

    def _check_path(self, path_to_check, need_write=False, can_create=False):
        if not path_to_check:
            return
        if not os.path.exists(path_to_check):
            if not can_create:
                raise BuildError('Missing required path: '
                                 '{}'.format(path_to_check))
            try:
                path_to_check.makedirs_p()
            except Exception:
                raise BuildError('Unable to create required path: '
                                 '{}'.format(path_to_check))
        if not os.access(path_to_check, os.R_OK):
            raise BuildError('Unable to read from: {}'.format(path_to_check))
        if need_write and not os.access(path_to_check, os.W_OK):
            raise BuildError('Unable to write to: {}'.format(path_to_check))

    def check_paths(self):
        self._check_path(self.build_dir, need_write=True, can_create=True)
        self._check_path(self.cache_dir, need_write=True, can_create=True)
        self._check_path(self.wheelhouse_overrides)

    def clean_removed(self, signatures):
        """
        Clean up any files that were accounted for in the previous build
        manifest but which have been removed in the current set of sigs.
        """
        old_sigs = json.loads(self.manifest.text())['signatures']
        old_files = set(old_sigs.keys()) - {'.build.manifest'}
        new_files = set(signatures.keys())
        removed = old_files - new_files
        for filename in removed:
            filepath = self.target_dir / filename
            filepath.remove_p()
        return removed

    def write_report(self, new_repo, added, changed, removed):
        """
        Log messages indicating what changed with this (re)build.
        """
        log.info('')
        log.info('---------------------------------------')
        log.info('              Build Report')
        log.info('---------------------------------------')

        if new_repo:
            log.info('New build; all files were modified.')
            return
        elif any([added, changed, removed]):
            sigils = ['+', ' ', '-']
            for sigil, filenames in zip(sigils, [added, changed, removed]):
                for filename in filenames:
                    log.info(' {} {}'.format(sigil, filename))
        else:
            log.info('No new changes; no files were modified.')

    def cleanup(self):
        log.debug('Cleaning up {}'.format(self.cache_dir))
        self.cache_dir.rmtree_p()


def make_url_from_lock_for_layer(lock_spec, use_branches=False):
    """Make a url from a lock spec for a layer or interface.

    lock_spec is:

    "layer:basic": {
      "branch": "refs/heads/master",
      "commit": "623e69c7b432456fd4364f6e1835424fd6b5425e",
      "item": "layer:basic",
      "type": "layer",
      "url": "https://github.com/juju-solutions/layer-basic.git",
      "vcs": null
    }

    It is assumed that the VCS is git.

    TODO: Add other VCS in addition to Git?

    :param lock_spec: the lock specification for the layer
    :type lock_spec: Dict[str, Dict[str, Optional[str]]]
    :param use_branches: if True, use the branch, rather than the commit
    :type use_branches: bool
    :returns: the url for fetching the layer from the repository
    :rtype: str
    """
    if use_branches:
        branch = lock_spec["branch"]
        if branch.startswith("refs/heads/"):
            branch = branch[len("refs/heads/"):]
        return "{}@{}".format(lock_spec["url"], branch)
    else:
        return "{}@{}".format(lock_spec["url"], lock_spec["commit"])


def configLogging(build):
    global log
    logging.captureWarnings(True)
    term = os.environ.get('TERM')
    if term and term.startswith('screen.'):
        term = term[7:]
    record_format = '%(name)s: %(message)s'
    try:
        clifmt = utils.ColoredFormatter(
            blessings.Terminal(term, force_styling=build.force_color),
            record_format)
    except curses.error:
        try:
            # try falling back to basic term type
            clifmt = utils.ColoredFormatter(
                blessings.Terminal('linux', force_styling=build.force_color),
                record_format)
        except curses.error:
            # fall back to uncolored formatter
            clifmt = logging.Formatter(record_format)
    root_logger = logging.getLogger()
    clihandler = logging.StreamHandler(sys.stdout)
    clihandler.setFormatter(clifmt)
    if isinstance(build.log_level, str):
        build.log_level = build.log_level.upper()
    root_logger.setLevel(build.log_level)
    log.setLevel(build.log_level)
    root_logger.addHandler(clihandler)
    requests_logger = logging.getLogger("requests")
    requests_logger.setLevel(logging.WARN)
    urllib_logger = logging.getLogger("urllib3")
    urllib_logger.setLevel(logging.CRITICAL)


def inspect(args=None):
    build = Builder()
    parser = argparse.ArgumentParser(
        description='inspect the layers of a built charm. This '
                    'renders a tree of the files in the color '
                    'of each layer. Each layers assigned color '
                    'is presented in a legend at the top of the '
                    'output.')
    parser.add_argument('-r', '--force-raw', action="store_true",
                        dest='force_color', help="Alias of --force-color")
    parser.add_argument('-c', '--force-color', action="store_true",
                        help="Force raw output (color)")
    parser.add_argument('-a', '--annotate', action="store_true",
                        help="Annotate each file with the layer name, "
                             "rather than just using colors "
                             "(this is enabled automatically enabled if "
                             "colors are unavailable)")
    parser.add_argument('-l', '--log-level', default=logging.INFO)
    parser.add_argument('charm', nargs="?", default=".", type=path)
    utils.add_plugin_description(parser)
    # Namespace will set the options as attrs of build
    parser.parse_args(args, namespace=build)
    configLogging(build)
    try:
        build.inspect()
    except BuildError as e:
        if e.args:
            log.error(*e.args)
        raise SystemExit(1)


def deprecated_main():
    namemap = {
        'compose': 'build',
        'generate': 'build',
        'refresh': 'build',
        'inspect': 'layers',
    }
    cmd = sys.argv[0]
    if "-" in cmd:
        old = cmd.rsplit('-', 1)[-1]
    else:
        old = sys.argv[1]
    new = namemap[old]

    class MockBuild(object):
        log_level = 'INFO'

    configLogging(MockBuild)

    msg = "{} has been deprecated, please use {}".format(old, new)
    if '--description' in sys.argv:
        print(msg)
    else:
        log.critical(msg)


def main(args=None):
    env_vars = (
        ("CHARM_LAYERS_DIR", "Directory containing local copies of base "
                             "layers"),
        ("CHARM_INTERFACES_DIR", "Directory containing local copies of "
                                 "interface layers"),
        ("CHARM_BUILD_DIR", "Build charms will be placed into "
                            "$CHARM_BUILD_DIR/{charm_name} "
                            "(defaults to /tmp/charm-builds/)"),
        ("JUJU_REPOSITORY", "Deprecated: If CHARM_BUILD_DIR is not set but "
                            "this is, built charms will be placed into "
                            "$JUJU_REPOSITORY/builds/{charm_name}"),
        ("LAYER_PATH", "Deprecated: Alias of CHARM_LAYERS_DIR"),
        ("INTERFACE_PATH", "Deprecated: Alias of CHARM_INTERFACES_DIR"),
    )
    build = Builder()
    parser = argparse.ArgumentParser(
        description="build a charm from layers and interfaces",
        epilog="The following environment variables will be honored:\n" +
               "\n" +
               "".join("  {:<20}  {}\n".format(name, desc)
                       for name, desc in env_vars),
        formatter_class=argparse.RawDescriptionHelpFormatter,)
    parser.add_argument('-l', '--log-level', default=logging.INFO)
    parser.add_argument('-f', '--force', action="store_true")
    parser.add_argument('-o', '--output-dir', type=path,
                        help='Alias for --build-dir')
    parser.add_argument('-d', '--build-dir', type=path,
                        help='Directory under which to place built charms; '
                             'overrides CHARM_BUILD_DIR env')
    parser.add_argument('-C', '--cache-dir', type=path,
                        help='Directory to cache build dependencies '
                        '(default: ~/.cache/charm; can also be set via '
                        'CHARM_CACHE_DIR env)')
    parser.add_argument('-s', '--series', default=None,
                        help='Deprecated: define series in metadata.yaml')
    parser.add_argument('--hide-metrics', dest="hide_metrics",
                        action="store_true")
    parser.add_argument('--interface-service',
                        help="Deprecated: use --layer-index")
    parser.add_argument('-i', '--layer-index',
                        help='One or more index URLs used to look up layers, '
                             'separated by commas. Can include the token '
                             'DEFAULT, which will be replaced by the default '
                             'index{} ({}).  E.g.: '
                             'https://my-site.com/index/,DEFAULT'.format(
                                 'es' if len(LayerFetcher.LAYER_INDEXES) > 1
                                 else '',
                                 ','.join(LayerFetcher.LAYER_INDEXES)))
    parser.add_argument('--no-local-layers', action="store_true",
                        help="Don't use local layers when building. "
                        "Forces included layers to be downloaded "
                        "from the interface service.")
    parser.add_argument('-n', '--name',
                        help="Build a charm of 'name' from 'charm'")
    parser.add_argument('--write-lock-file', action="store_true",
                        default=False,
                        help="Write a lock file for reproducible builds. The "
                             "file is 'layers.lock' in the root of the layer "
                             "being build.")
    parser.add_argument('--use-lock-file-branches', action="store_true",
                        default=False,
                        help="Use branch names if not master branch in "
                             "lockfile. This allows tracking of a stable "
                             "branch.")
    parser.add_argument('--ignore-lock-file', action="store_true",
                        default=False,
                        help="Ignore the lock file even if it is present.")
    parser.add_argument('-r', '--report', action="store_true", default=True,
                        help="Show post-build report of changes")
    parser.add_argument('-R', '--no-report', action="store_false",
                        dest='report', default=True,
                        help="Don't show post-build report of changes")
    parser.add_argument('-w', '--wheelhouse-overrides', type=path,
                        help="Provide a wheelhouse.txt file with overrides "
                             "for the built wheelhouse")
    parser.add_argument('-W', '--wheelhouse-per-layer', action="store_true",
                        help="Deprecated: Use original wheelhouse processing "
                             "method (see PR juju/charm-tools#569)")
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help="Increase output (same as -l DEBUG)")
    parser.add_argument('--debug', action='store_true',
                        help='Same as --log-level=DEBUG')
    parser.add_argument('-c', '--force-color', action="store_true",
                        help="Force raw output (color)")
    parser.add_argument('--charm-file', '-F', action='store_true',
                        help='Create a .charm file in the current directory')
    parser.add_argument('charm', nargs="?", default=".", type=path,
                        help='Source directory for charm layer to build '
                             '(default: .)')
    utils.add_plugin_description(parser)
    # Namespace will set the options as attrs of build
    parser.parse_args(args, namespace=build)
    if build.charm == "help":
        parser.print_help()
        raise SystemExit(0)

    if build.verbose or build.debug:
        build.log_level = logging.DEBUG

    LayerFetcher.set_layer_indexes(build.interface_service or
                                   build.layer_index)
    LayerFetcher.NO_LOCAL_LAYERS = build.no_local_layers

    WheelhouseTactic.per_layer = build.wheelhouse_per_layer

    configLogging(build)

    try:
        build.check_series()
        build.normalize_build_dir()
        build.normalize_cache_dir()
        build.check_paths()
        build.maybe_read_lock_file()

        build()

        lint, exit_code = proof.proof(build.target_dir, False, False)
        llog = logging.getLogger("proof")

        if not lint:
            llog.info('OK!')

        for line in lint:
            if line[0] == "I":
                llog.info(line)
            elif line[0] == "W":
                llog.warn(line)
            elif line[0] == "E":
                llog.error(line)

        if exit_code > 0:
            raise SystemExit(exit_code)

    except (BuildError, FetchError) as e:
        log.debug(traceback.format_exc())
        if e.args:
            log.error(*e.args)
        raise SystemExit(1)


if __name__ == '__main__':
    main()



from .tactics import load_tactic


from ruamel import yaml
import logging
from path import Path as path
from otherstuf import chainstuf

DEFAULT_IGNORES = [
    ".bzr",
    ".git",
    "**/.ropeproject",
    "*.pyc",
    "*~",
    ".tox",
    "*.swp",
    ".pull-source-rev",
    "*.charm",
]


class BuildConfig(chainstuf):
    """Defaults for controlling the generator, each layer in
    the inclusion graph can provide values, including things
    like overrides, or warnings if things are overridden that
    shouldn't be.
    """

    def __init__(self, *args, **kwargs):
        super(BuildConfig, self).__init__(*args, **kwargs)
        self['_tactics'] = self.get('tactics', [])
        self.configured = False

    def __getattr__(self, key):
        return self[key]

    def rget(self, key):
        """Combine all the results from all the layers into a single iter"""
        result = []
        for m in self.maps:
            r = m.get(key)
            if r:
                if isinstance(r, (list, tuple)):
                    result.extend(r)
                else:
                    result.append(r)
        return result

    def configure(self, config_file, allow_missing=False):
        config_file = path(config_file)
        data = None
        if not config_file.exists() and not allow_missing:
            raise OSError("Missing Config File {}".format(config_file))
        try:
            if config_file.exists() and config_file.text('utf-8').strip():
                data = yaml.safe_load(config_file.open(encoding='utf-8'))
                self.configured = True
        except yaml.error.YAMLError as e:
            logging.critical("Malformed config file {}: {}".format(config_file,
                                                                   e))
            raise
        if data:
            self.update(data)
        # look at any possible imports and use them to build tactics
        tactics = self.get('tactics')
        basedir = config_file.dirname()
        if tactics:
            for name in tactics:
                tactic = load_tactic(name, basedir)
                self._tactics.append(tactic)
        return self

    @classmethod
    def from_config(cls, config_file, allow_missing=False):
        c = cls()
        c.configure(config_file, allow_missing)
        return c

    def add_config(self, config):
        new_config = self.new_child()
        new_config.update(config)
        return new_config

    @property
    def name(self):
        return self.get('name')

    @property
    def tactics(self):
        return self.rget('_tactics')

    @property
    def ignores(self):
        """
        The ignores for this layer (lower layers are not included).
        """
        ignores = list(DEFAULT_IGNORES)
        if self.maps:
            ignores.extend(self.maps[0].get('ignore', []))
        return ignores

    @property
    def excludes(self):
        """
        The excludes for this layer (lower layers are not included).
        """
        excludes = list(DEFAULT_IGNORES)
        if self.maps:
            excludes.extend(self.maps[0].get('exclude', []))
        return excludes



import os
import json
import logging
import shutil

import requests
from charmtools import fetchers
from charmtools.fetchers import (git,  # noqa
                                 Fetcher,
                                 get_fetcher,
                                 FetchError)

from path import Path as path


log = logging.getLogger(__name__)


class RepoFetcher(fetchers.LocalFetcher):
    @classmethod
    def can_fetch(cls, url):
        search_path = [os.getcwd(), os.environ.get("JUJU_REPOSITORY", ".")]
        cp = os.environ.get("LAYER_PATH")
        if cp:
            search_path.extend(cp.split(":"))
        for part in search_path:
            p = (path(part) / url).normpath()
            if p.exists():
                return dict(path=p)
        return {}


fetchers.FETCHERS.insert(0, RepoFetcher)


class LayerFetcher(fetchers.LocalFetcher):
    _DEFAULT_LAYER_INDEXES = ["https://juju.github.io/layer-index/"]
    LAYER_INDEXES = _DEFAULT_LAYER_INDEXES
    NO_LOCAL_LAYERS = False
    NAMESPACE = "layer"
    ENVIRON = "CHARM_LAYERS_DIR"
    OLD_ENVIRON = "LAYER_PATH"
    OPTIONAL_PREFIX = "juju-layer-"
    ENDPOINT = "layers"
    _DEFAULT_BRANCH = None
    BRANCH = _DEFAULT_BRANCH

    @classmethod
    def set_layer_indexes(cls, layer_indexes):
        if not layer_indexes:
            return
        if isinstance(layer_indexes, str):
            layer_indexes = layer_indexes.split(',')
        new_indexes = []
        for layer_index in layer_indexes:
            if layer_index == 'DEFAULT':
                new_indexes.extend(cls._DEFAULT_LAYER_INDEXES)
            else:
                new_indexes.append(layer_index)
        cls.LAYER_INDEXES = new_indexes

    @classmethod
    def restore_layer_indexes(cls):
        cls.LAYER_INDEXES = cls._DEFAULT_LAYER_INDEXES

    @classmethod
    def set_branch(cls, branch):
        if not branch:
            return
        cls.BRANCH = branch

    @classmethod
    def restore_branch(cls):
        cls.BRANCH = cls._DEFAULT_BRANCH

    @classmethod
    def can_fetch(cls, url):
        # Search local path first, then the interface webservice
        if url.startswith("{}:".format(cls.NAMESPACE)):
            name = url[len(cls.NAMESPACE) + 1:]

            if not cls.NO_LOCAL_LAYERS:
                prefixed_name = '{}-{}'.format(cls.NAMESPACE, name)
                search_path = []
                if cls.ENVIRON in os.environ:
                    search_path.append(os.environ[cls.ENVIRON])
                elif cls.OLD_ENVIRON in os.environ:
                    search_path.append(os.environ[cls.OLD_ENVIRON])
                else:
                    search_path.append(os.environ.get("JUJU_REPOSITORY", "."))
                for part in search_path:
                    basepath = path(part)
                    for dirname in (name, prefixed_name):
                        p = (basepath / dirname).normpath()
                        if p.exists():
                            return dict(path=p)

            choices = [name]
            if name.startswith(cls.OPTIONAL_PREFIX):
                choices.append(name[len(cls.OPTIONAL_PREFIX):])
            for choice in choices:
                for layer_index in cls.LAYER_INDEXES:
                    uri = "%s%s/%s.json" % (
                        layer_index, cls.ENDPOINT, choice)
                    log.debug('Checking layer index: {}'.format(uri))
                    if uri.startswith('file://'):
                        choice_path = path(uri[7:])
                        if not choice_path.exists():
                            continue
                        try:
                            result = json.loads(choice_path.text())
                        except json.JSONDecodeError as e:
                            log.error('Unable to parse index entry for {}: '
                                      '{}'.format(url, e))
                            continue
                        if not result.get('repo'):
                            continue
                        log.debug('Found repo: {}'.format(result['repo']))
                        return result
                    try:
                        result = requests.get(uri)
                    except Exception:
                        result = None
                    if result and result.ok:
                        result = result.json()
                        if result.get("repo"):
                            log.debug('Found repo: {}'.format(result['repo']))
                            return result
            return {}

    def target(self, dir_):
        """Return a :class:`path` of the directory where the downloaded item
        will be located.

        :param str dir_: Directory into which the item will be downloaded.
        :return: :class:`path`

        """
        if hasattr(self, "path"):
            return self.path
        elif hasattr(self, "repo"):
            _, target = self._get_repo_fetcher_and_target(self.repo, dir_)
            return target

    def _get_repo_fetcher_and_target(self, repo, dir_):
        """Returns a :class:`Fetcher` for ``repo``, and the destination dir
        at which the downloaded repo will be created.

        :param str repo: The repo url.
        :param str dir_: Directory into which the repo will be downloaded.
        :return: 2-tuple of (:class:`Fetcher`, :class:`path`)

        """
        u = self.url[len(self.NAMESPACE) + 1:]
        f = get_fetcher(repo)
        return f, path(dir_) / u

    def fetch(self, dir_):
        if hasattr(self, "path"):
            return super(LayerFetcher, self).fetch(dir_)
        elif hasattr(self, "repo"):
            f, target = self._get_repo_fetcher_and_target(self.repo, dir_)
            log.debug('Using fetcher: {}'.format(f))
            if self.BRANCH is not None:
                log.debug('Adding branch: %s', self.BRANCH)
                f.revision = self.BRANCH
            orig_res = res = f.fetch(dir_)
            log.debug("url fetched (for lockfile): %s",
                      getattr(f, 'fetched_url'))
            self.fetched_url = getattr(f, 'fetched_url', None)
            self.vcs = getattr(f, 'vcs', None)
            # make sure we save the revision of the actual repo, before we
            # start traversing subdirectories and moving contents around
            self.revision = self.get_revision(res)
            if res != target:
                res = path(res)
                if hasattr(self, 'subdir'):
                    res = res / self.subdir
                target.rmtree_p()
                log.debug('Copying {} to {}'.format(res, target))
                shutil.copytree(res, target)
                log.debug('Cleaning up {}'.format(orig_res))
                path(orig_res).rmtree_p()  # cleanup from cache after copied
            return target


fetchers.FETCHERS.insert(0, LayerFetcher)


class InterfaceFetcher(LayerFetcher):
    NAMESPACE = "interface"
    ENVIRON = "CHARM_INTERFACES_DIR"
    OLD_ENVIRON = "INTERFACE_PATH"
    OPTIONAL_PREFIX = "juju-relation-"
    ENDPOINT = "interfaces"


fetchers.FETCHERS.insert(0, InterfaceFetcher)



import errno
import logging
import os
import re
import shlex
import shutil
import subprocess
import tempfile

import requests
import yaml
from path import Path as path


log = logging.getLogger(__name__)

REQUEST_TIMEOUT_SECS = 45


def get(*args, **kw):
    if 'timeout' not in kw:
        kw['timeout'] = REQUEST_TIMEOUT_SECS

    return requests.get(*args, **kw)


def is_int(string):
    try:
        int(string)
        return True
    except ValueError:
        return False


def rename(dir_):
    """If ``dir_`` is a charm directory, rename it to match the
    charm name, otherwise do nothing.

    :param dir_: directory path
    :return: the new directory name (possibly unchanged).

    """
    dir_ = dir_.rstrip(os.sep)
    metadata = os.path.join(dir_, "metadata.yaml")
    if not os.path.exists(metadata):
        return dir_
    metadata = yaml.safe_load(open(metadata))
    if not metadata:
        return dir_
    name = metadata.get("name")
    if not name:
        return dir_
    new_dir = os.path.join(os.path.dirname(dir_), name)
    if not os.path.exists(new_dir):
        # This ignores existing repos
        # In truth we want control over management of existing
        # repos with per VCS branch selections
        # ex: switching the git branch marked in revision
        os.rename(dir_, new_dir)
    return new_dir


def extract_archive(archive, dir_):
    """Extract zip archive at filesystem path ``archive`` into directory
    ``dir_`` and return the full path to the directory containing the
    extracted archive.

    """
    tempdir = tempfile.mkdtemp(dir=dir_)
    log.debug("Extracting %s to %s", archive, tempdir)
    # Can't  extract with python due to bug that drops file
    # permissions: http://bugs.python.org/issue15795
    # In particular, it's important that executable test files in the
    # archive remain executable, otherwise the tests won't be run.
    # Instead we use a shell equivalent of the following:
    #     archive = zipfile.ZipFile(archive, 'r')
    #     archive.extractall(tempdir)
    check_call('unzip {} -d {}'.format(archive, tempdir))
    return tempdir


def download_file(url, dir_):
    """Download file at ``url`` into directory ``dir_`` and return the full
    path to the downloaded file.

    """
    _, filename = tempfile.mkstemp(dir=dir_)
    log.debug("Downloading %s", url)
    r = get(url, stream=True)
    with open(filename, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
                f.flush()
    return filename


class Fetcher(object):
    def __init__(self, url, **kw):
        self.revision = ''
        self.url = url
        for k, v in kw.items():
            setattr(self, k, v)

    @classmethod
    def can_fetch(cls, url):
        match = cls.MATCH.search(url)
        return match.groupdict() if match else {}

    def get_revision(self, dir_):
        for cmd in ("git rev-parse HEAD",
                    "bzr revision-info",
                    "hg log -l 1 --template '{node}\n' -r ."):
            try:
                rev_info = check_output(cmd, cwd=dir_)
                if cmd.startswith('git'):
                    return rev_info.decode('utf8').strip()
                if cmd.startswith('bzr'):
                    return rev_info.decode('utf8').strip().split()[1]
                if cmd.startswith('hg'):
                    return rev_info.decode('utf8').strip()
            except FetchError:
                continue
        rev_file = path(dir_) / '.pull-source-rev'
        if rev_file.exists():
            return rev_file.read_text().strip()
        return self.revision

    def get_branch_for_revision(self, dir_, revision=None):
        """Returns None if the revision doesn't match the actual branch name"""
        if revision is None:
            revision = self.revision
        if not revision:
            return None
        for cmd in ("git branch --contains {} --format=\"%(refname)\""
                    .format(revision), ):
            try:
                branch = check_output(cmd, cwd=dir_).decode('UTF-8').strip()
                return branch
            except FetchError:
                continue
        return None


class BzrFetcher(Fetcher):
    MATCH = re.compile(r"""
    ^(lp:|launchpad:|https?://((code|www)\.)?launchpad.net/|bzr\+ssh://[^/]+/)
    (?P<repo>[^@]*)(@(?P<revision>.*))?$
    """, re.VERBOSE)

    @classmethod
    def can_fetch(cls, url):
        matchdict = super(BzrFetcher, cls).can_fetch(url)
        return matchdict if '/+merge/' not in matchdict.get('repo', '') else {}

    def fetch(self, dir_):
        dir_ = tempfile.mkdtemp(dir=dir_)
        url = 'lp:' + self.repo
        self.fetched_url = url
        self.vcs = "bzr"
        cmd = 'branch --use-existing-dir {} {}'.format(url, dir_)
        if self.revision:
            log.debug('Switching to revision: {}'.format(self.revision))
            cmd = '{} -r {}'.format(cmd, self.revision)
        bzr(cmd)
        return rename(dir_)


class BzrMergeProposalFetcher(BzrFetcher):
    @classmethod
    def can_fetch(cls, url):
        matchdict = super(BzrFetcher, cls).can_fetch(url)
        return matchdict if '/+merge/' in matchdict.get('repo', '') else {}

    def fetch(self, dir_):
        dir_ = tempfile.mkdtemp(dir=dir_)
        api_base = 'https://api.launchpad.net/devel/'
        url = api_base + self.repo
        self.fetched_url = url
        self.vcs = "bzr"
        merge_data = get(url).json()
        target = 'lp:' + merge_data['target_branch_link'][len(api_base):]
        source = 'lp:' + merge_data['source_branch_link'][len(api_base):]
        bzr('branch --use-existing-dir {} {}'.format(target, dir_))
        bzr('merge {}'.format(source), cwd=dir_)
        bzr('commit --unchanged -m "Merge commit"', cwd=dir_)
        return rename(dir_)


class LaunchpadGitFetcher(Fetcher):
    MATCH = re.compile(r"""
    ^(git:|https)?://git.launchpad.net/
    (?P<repo>[^@]*)(@(?P<revision>.*))?$
    """, re.VERBOSE)

    def fetch(self, dir_):
        dir_ = tempfile.mkdtemp(dir=dir_)
        url = 'https://git.launchpad.net/' + self.repo
        self.fetched_url = url
        self.vcs = "git"
        git('clone {} {}'.format(url, dir_))
        if self.revision:
            log.debug('Switching to revision: {}'.format(self.revision))
            git('checkout {}'.format(self.revision), cwd=dir_)
        return rename(dir_)


class GithubFetcher(Fetcher):
    MATCH = re.compile(r"""
    ^(gh:|github:|https?://(www\.)?github.com/|git@github.com:)
    (?P<repo>[^@]*)(@(?P<revision>.*))?$
    """, re.VERBOSE)

    def fetch(self, dir_):
        dir_ = tempfile.mkdtemp(dir=dir_)
        url = 'https://github.com/' + self.repo
        self.fetched_url = url
        self.vcs = "git"
        git('clone {} {}'.format(url, dir_))
        if self.revision:
            log.debug('Switching to revision: {}'.format(self.revision))
            git('checkout {}'.format(self.revision), cwd=dir_)
        return rename(dir_)


class OpendevFetcher(Fetcher):
    MATCH = re.compile(r"""
    ^(https:\/\/(www\.)?opendev\.org\/|git@opendev.org:)
    (?P<repo>[^@]*)(@(?P<revision>.*))?$
    """, re.VERBOSE)

    def fetch(self, dir_):
        dir_ = tempfile.mkdtemp(dir=dir_)
        url = 'https://opendev.org/' + self.repo
        self.fetched_url = url
        self.vcs = "git"
        git('clone {} {}'.format(url, dir_))
        if self.revision:
            log.debug('Switching to revision: {}'.format(self.revision))
            git('checkout {}'.format(self.revision), cwd=dir_)
        return rename(dir_)


class GitFetcher(Fetcher):
    """Generic git fetcher.

    Matches any url that starts with "git" or ends with ".git".

    """
    MATCH = re.compile(r"""
    ^(?P<repo>git.*|.*\.git)?$
    """, re.VERBOSE)

    def fetch(self, dir_):
        dir_ = tempfile.mkdtemp(dir=dir_)
        self.fetched_url = self.repo
        self.vcs = "git"
        git('clone {} {}'.format(self.repo, dir_))
        if self.revision:
            log.debug('Switching to revision: {}'.format(self.revision))
            git('checkout {}'.format(self.revision), cwd=dir_)
        return rename(dir_)


class BitbucketFetcher(Fetcher):
    MATCH = re.compile(r"""
    ^(bb:|bitbucket:|https?://(www\.)?bitbucket.org/)
    (?P<repo>[^@]*)(@(?P<revision>.*))?$
    """, re.VERBOSE)

    def fetch(self, dir_):
        dir_ = tempfile.mkdtemp(dir=dir_)
        url = 'https://bitbucket.org/' + self.repo
        self.fetched_url = url
        if url.endswith('.git'):
            self.vcs = "git"
            return self._fetch_git(url, dir_)
        self.vcs = "hg"
        return self._fetch_hg(url, dir_)

    def _fetch_git(self, url, dir_):
        git('clone {} {}'.format(url, dir_))
        if self.revision:
            log.debug('Switching to revision: {}'.format(self.revision))
            git('checkout {}'.format(self.revision), cwd=dir_)
        return rename(dir_)

    def _fetch_hg(self, url, dir_):
        cmd = 'clone {} {}'.format(url, dir_)
        if self.revision:
            log.debug('Switching to revision: {}'.format(self.revision))
            cmd = '{} -u {}'.format(cmd, self.revision)
        hg(cmd)
        return rename(dir_)


class LocalFetcher(Fetcher):
    @classmethod
    def can_fetch(cls, url):
        src = os.path.abspath(
            os.path.join(os.getcwd(), os.path.expanduser(url)))
        if os.path.exists(src):
            return dict(path=src)
        return {}

    def fetch(self, dir_):
        dst = os.path.join(dir_, os.path.basename(self.path.rstrip(os.sep)))
        shutil.copytree(self.path, dst, symlinks=True)
        return dst


class CharmstoreDownloader(Fetcher):
    """Downloads and extracts a charm archive from the charm store.

    """
    MATCH = re.compile(r"""
    ^cs:(?P<entity>.*)$
    """, re.VERBOSE)

    STORE_URL = 'https://api.jujucharms.com/charmstore/v4/{}'
    ARCHIVE_URL = STORE_URL + '/archive'
    REVISION_URL = STORE_URL + '/meta/id-revision'

    def __init__(self, *args, **kw):
        super(CharmstoreDownloader, self).__init__(*args, **kw)

    def fetch(self, dir_):
        url = self.ARCHIVE_URL.format(self.entity)
        self.fetched_url = url
        self.vcs = "charmstore"
        archive = download_file(url, dir_)
        entity_dir = extract_archive(archive, dir_)
        return rename(entity_dir)

    def get_revision(self, dir_):
        url = self.REVISION_URL.format(self.entity)
        return get(url).json()['Revision']


class BundleDownloader(CharmstoreDownloader):
    MATCH = re.compile(r"""
    ^bundle:(?P<entity>.*)$
    """, re.VERBOSE)

    def __init__(self, *args, **kw):
        super(BundleDownloader, self).__init__(*args, **kw)
        self.entity = normalize_bundle_name(self.entity)


def normalize_bundle_name(bundle_name):
    """Convert old-style bundle name to new format.

    Example:
        ~charmers/mediawiki/6/single -> ~charmers/mediawiki-single-6

    (for more examples see tests)

    """
    owner, bundle = None, bundle_name
    if bundle.startswith('~'):
        owner, bundle = bundle.split('/', 1)
    bundle_parts = bundle.split('/')
    if len(bundle_parts) == 3 and is_int(bundle_parts[1]):
        bundle_parts = [
            bundle_parts[0],
            bundle_parts[2],
            bundle_parts[1]]
    bundle = '-'.join(bundle_parts)
    if owner:
        bundle = '/'.join((owner, bundle))
    return bundle


def bzr(cmd, **kw):
    check_call('bzr ' + cmd, **kw)


def git(cmd, **kw):
    check_call('git ' + cmd, **kw)


def hg(cmd, **kw):
    check_call('hg ' + cmd, **kw)


def check_call(cmd, **kw):
    return check_output(cmd, **kw)


class FetchError(Exception):
    pass


def check_output(cmd, **kw):
    args = shlex.split(cmd)
    try:
        p = subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            **kw
        )
    except OSError as e:
        msg = 'Unable to run "%s": %s' % (args[0], e.strerror)
        if e.errno == errno.ENOENT:
            msg += '\nPlease install "%s" and try again' % args[0]
        raise FetchError(msg)
    out, _ = p.communicate()
    if p.returncode != 0:
        raise FetchError(out.decode('utf8'))
    log.debug('%s: %s', cmd, out)
    return out


FETCHERS = [
    BzrFetcher,
    BzrMergeProposalFetcher,
    GithubFetcher,
    OpendevFetcher,
    BitbucketFetcher,
    LocalFetcher,
    CharmstoreDownloader,
    BundleDownloader,
    LaunchpadGitFetcher,
    GitFetcher,
]


def get_fetcher(url):
    for fetcher in FETCHERS:
        matchdict = fetcher.can_fetch(url)
        if matchdict:
            return fetcher(url, **matchdict)
    raise FetchError('No fetcher for url: %s' % url)



#!/usr/bin/python
#
#    pull-source - Fetch source for charm, layers, and interfaces
#
#    Copyright (C) 2016  Canonical Ltd.
#    Author: Tim Van Steenburgh <tvansteenburgh@gmail.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""download the source code for a charm, layer, or interface.

The item to download can be specified using any of the following forms:

 - [cs:]charm
 - [cs:]series/charm
 - [cs:]~user/charm
 - [cs:]~user/series/charm
 - layer:layer-name
 - interface:interface-name

If the item is a layered charm, and the top layer of the charm has a repo
key in layer.yaml, the top layer repo will be cloned. Otherwise, the charm
archive will be downloaded and extracted from the charm store.

If a download directory is not specified, the following environment vars
will be used to determine the download location:

 - For charms, $CHARM_BUILD_DIR/../charms or $JUJU_REPOSITORY (deprecated)
 - For layers, $CHARM_LAYERS_DIR or $LAYER_PATH (deprecated)
 - For interfaces, $CHARM_INTERFACES_DIR or $INTERFACE_PATH (deprecated)

If a download location can not be determined from environment variables,
the current working directory will be used.

The download is aborted if the destination directory already exists.

"""

import argparse
import atexit
import logging
import os
import shutil
import sys
import tempfile
import textwrap

import yaml
from path import Path as path

from charmtools import utils
from charmtools.build import fetchers
from charmtools.fetchers import (
    CharmstoreDownloader,
    FETCHERS,
    get,
)


log = logging.getLogger(__name__)

LAYER_PREFIX = fetchers.LayerFetcher.NAMESPACE + ':'
INTERFACE_PREFIX = fetchers.InterfaceFetcher.NAMESPACE + ':'
CHARM_PREFIX = 'cs:'
CHARM_LAYERS_DIR = os.environ.get(
    fetchers.LayerFetcher.ENVIRON,
    os.environ.get(fetchers.LayerFetcher.OLD_ENVIRON))
CHARM_INTERFACES_DIR = os.environ.get(
    fetchers.InterfaceFetcher.ENVIRON,
    os.environ.get(fetchers.InterfaceFetcher.OLD_ENVIRON))
if 'CHARM_BUILD_DIR' in os.environ:
    CHARM_CHARMS_DIR = os.path.join(
        os.path.dirname(os.environ['CHARM_BUILD_DIR']), 'charms')
else:
    CHARM_CHARMS_DIR = os.environ.get('JUJU_REPOSITORY')

ERR_DIR_EXISTS = "Aborting, destination directory exists"


class CharmstoreRepoDownloader(CharmstoreDownloader):
    """Clones a charm's bzr repo.

    If the a bzr repo is not set, falls back to
    :class:`fetchers.CharmstoreDownloader`.

    """
    EXTRA_INFO_URL = CharmstoreDownloader.STORE_URL + '/meta/extra-info'

    def fetch(self, dir_):
        url = self.EXTRA_INFO_URL.format(self.entity)
        repo_url = get(url).json().get('bzr-url')
        if repo_url:
            try:
                fetcher = fetchers.get_fetcher(repo_url)
            except fetchers.FetchError:
                log.debug(
                    "No fetcher for %s, downloading from charmstore",
                    repo_url)
                return super(CharmstoreRepoDownloader, self).fetch(dir_)
            else:
                return fetcher.fetch(dir_)
        return super(CharmstoreRepoDownloader, self).fetch(dir_)


FETCHERS.insert(0, CharmstoreRepoDownloader)


class CharmstoreLayerDownloader(CharmstoreRepoDownloader):
    """Clones the repo containing the top layer of a charm.

    If the charm is not a layered charm, or the repo for the
    top layer can not be determined, falls back to using
    :class:`CharmstoreRepoDownloader`.

    """
    LAYER_CONFIGS = ['layer.yaml', 'composer.yaml']

    def fetch(self, dir_):
        for cfg in self.LAYER_CONFIGS:
            url = '{}/{}'.format(
                self.ARCHIVE_URL.format(self.entity), cfg)
            result = get(url)
            if not result.ok:
                continue
            repo_url = yaml.safe_load(result.text).get('repo')
            if not repo_url:
                continue
            try:
                fetcher = fetchers.get_fetcher(repo_url)
            except fetchers.FetchError:
                log.debug(
                    'Charm %s has a repo set in %s, but no fetcher could '
                    'be found for the repo (%s).', self.entity, cfg, repo_url)
                break
            else:
                return fetcher.fetch(dir_)
        return super(CharmstoreLayerDownloader, self).fetch(dir_)


FETCHERS.insert(0, CharmstoreLayerDownloader)


def download_item(args):
    series_dir = None

    if args.item.startswith(LAYER_PREFIX):
        dir_ = args.dir or CHARM_LAYERS_DIR
        name = args.item[len(LAYER_PREFIX):]
    elif args.item.startswith(INTERFACE_PREFIX):
        dir_ = args.dir or CHARM_INTERFACES_DIR
        name = args.item[len(INTERFACE_PREFIX):]
    else:
        dir_ = args.dir or CHARM_CHARMS_DIR
        if not args.item.startswith(CHARM_PREFIX):
            args.item = CHARM_PREFIX + args.item

        url_parts = args.item[len(CHARM_PREFIX):].split('/')
        name = url_parts[-1]
        if len(url_parts) == 2 and not url_parts[0].startswith('~'):
            series_dir = url_parts[0]
        elif len(url_parts) == 3:
            series_dir = url_parts[1]

    dir_ = dir_ or os.getcwd()
    dir_ = os.path.abspath(os.path.expanduser(dir_))

    if not os.access(dir_, os.W_OK):
        print('Unable to write to {}'.format(dir_))
        return 200

    # Create series dir if we need to
    if series_dir:
        series_path = os.path.join(dir_, series_dir)
        if not os.path.exists(series_path):
            os.mkdir(series_path)
        dir_ = series_path

    # Abort if destination dir already exists
    final_dest_dir = os.path.join(dir_, name)
    if os.path.exists(final_dest_dir):
        print("{}: {}".format(ERR_DIR_EXISTS, final_dest_dir))
        return 1

    # Create tempdir for initial download
    tempdir = tempfile.mkdtemp()
    atexit.register(shutil.rmtree, tempdir)
    try:
        # Download the item
        fetcher = fetchers.get_fetcher(args.item)
        download_dir = fetcher.fetch(tempdir)
    except fetchers.FetchError:
        print("Can't find source for {}".format(args.item))
        return 1

    # Copy download dir to final destination dir
    shutil.copytree(download_dir, final_dest_dir, symlinks=True)
    rev = ' (rev: {})'.format(fetcher.revision) if fetcher.revision else ''
    if fetcher.revision:
        rev_file = path(final_dest_dir) / '.pull-source-rev'
        rev_file.write_text(fetcher.revision)
    print('Downloaded {}{} to {}'.format(args.item, rev, final_dest_dir))


def setup_parser():
    parser = argparse.ArgumentParser(
        prog='charm pull-source',
        description=textwrap.dedent(__doc__),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        'item',
        help='Name of the charm, layer, or interface to download.'
    )
    parser.add_argument(
        'dir', nargs='?',
        help='Directory in which to place the downloaded source.',
    )
    parser.add_argument(
        '-b', '--branch',
        help='Branch to check out after cloning the repo '
             '(before copying any subdir). If not given, '
             'the default branch of the repo will be used.'
    )
    parser.add_argument(
        '-i', '--layer-index',
        help='One or more index URLs used to look up layers, '
             'separated by commas. Can include the token '
             'DEFAULT, which will be replaced by the default '
             'index{} ({}).  E.g.: https://my-site.com/index/,DEFAULT'.format(
                 'es' if len(fetchers.LayerFetcher.LAYER_INDEXES) > 1 else '',
                 ','.join(fetchers.LayerFetcher.LAYER_INDEXES)))
    parser.add_argument(
        '-v', '--verbose',
        help='Show verbose output',
        action='store_true', default=False,
    )
    utils.add_plugin_description(parser)

    return parser


def main():
    parser = setup_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(
            format='%(levelname)s %(filename)s: %(message)s',
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format='%(levelname)s: %(message)s',
            level=logging.WARN,
        )

    fetchers.LayerFetcher.NO_LOCAL_LAYERS = True
    fetchers.LayerFetcher.set_layer_indexes(args.layer_index)
    fetchers.LayerFetcher.set_branch(args.branch)

    return download_item(args)


if __name__ == "__main__":
    sys.exit(main())



import argparse
import copy
import collections
import hashlib
import importlib
import json
import logging
import os
import re
import six
import subprocess
import sys
import tempfile
import time
from contextlib import contextmanager

from .diff_match_patch import diff_match_patch
import blessings
import pathspec
from path import Path as path

log = logging.getLogger('utils')


@contextmanager
def cd(directory, make=False):
    cwd = os.getcwd()
    if not os.path.exists(directory) and make:
        os.makedirs(directory)
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(cwd)


@contextmanager
def tempdir(chdir=True):
    dirname = path(tempfile.mkdtemp())
    if chdir:
        with cd(dirname):
            yield dirname
    else:
        yield dirname
    dirname.rmtree_p()


def deepmerge(dest, src):
    """
    Deep merge all the keys in src. When the value of the key is a dict,
    recursively call deepmerge. When the value of the key is a list, iterate
    the entire list and de-duplicate the list. When neither a dict or a list
    call the python copy.deepcopy method.

    This is destructive (`dest` is modified), as values from `src` may be
    passed through `copy.deepcopy`.
    """
    for k, v in src.items():
        if dest.get(k) and isinstance(v, dict):
            deepmerge(dest[k], v)
        elif dest.get(k) and isinstance(v, list):
            for item in v:
                if item not in dest[k]:
                    dest[k].append(item)
        else:
            dest[k] = copy.deepcopy(v)
    return dest


def delete_path(path, obj):
    """Delete a dotted path from object, assuming each level is a dict"""
    # TODO: Support lists
    parts = path.split('.')
    for p in parts[:-1]:
        obj = obj[p]
    if parts[-1] in obj:
        del obj[parts[-1]]


class NestedDict(dict):
    def __init__(self, dict_or_iterable=None, **kwargs):
        if dict_or_iterable:
            if isinstance(dict_or_iterable, dict):
                self.update(dict_or_iterable)
            elif isinstance(dict_or_iterable, collections.Iterable):
                for k, v in dict_or_iterable:
                    self[k] = v
        if kwargs:
            self.update(kwargs)

    def __setitem__(self, key, value):
        key = key.split('.')
        o = self
        for part in key[:-1]:
            o = o.setdefault(part, self.__class__())
        dict.__setitem__(o, key[-1], value)

    def __getitem__(self, key):
        o = self
        if '.' in key:
            parts = key.split('.')
            key = parts[-1]
            for part in parts[:-1]:
                o = o[part]

        return dict.__getitem__(o, key)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def update(self, other):
        deepmerge(self, other)


class ProcessResult(object):
    def __init__(self, command, exit_code, stdout, stderr):
        self.command = command
        self.exit_code = exit_code
        self.stdout = stdout
        self.stderr = stderr

    def __repr__(self):
        return '<ProcessResult "%s" result %s>' % (self.cmd, self.exit_code)

    @property
    def cmd(self):
        return ' '.join(self.command)

    @property
    def output(self):
        result = ''
        if self.stdout:
            stdout = self.stdout
            if six.PY3:
                stdout = stdout.decode('UTF-8')
            result += stdout
        if self.stderr:
            stderr = self.stderr
            if six.PY3:
                stderr = stderr.decode('UTF-8')
            result += stderr
        return result.strip()

    @property
    def json(self):
        if self.stdout:
            return json.loads(self.stdout)
        return None

    def __eq__(self, other):
        return self.exit_code == other

    def __bool__(self):
        return self.exit_code == 0

    __nonzero__ = __bool__

    def exit_on_error(self):
        if not bool(self):
            sys.stderr.write(
                '{}\n\nCommand failed: {}\n'.format(self.output, self.cmd))
            sys.exit(self.exit_code)


class Process(object):
    def __init__(self, command=None, exit=False, log=log, **kwargs):
        if isinstance(command, str):
            command = (command, )
        self.command = command
        self._exit_on_error = exit
        self.log = log
        self._kw = kwargs

    def __repr__(self):
        return "<Command %s>" % (self.command, )

    def exit_on_error(self, exit=True):
        self._exit_on_error = exit
        return self

    def __call__(self, *args, **kw):
        kwargs = dict(stdout=subprocess.PIPE,
                      stderr=subprocess.STDOUT)
        if self._kw:
            kwargs.update(self._kw)
        kwargs.update(kw)
        if self.command:
            all_args = self.command + args
        else:
            all_args = args
        if 'env' not in kwargs:
            kwargs['env'] = os.environ

        p = subprocess.Popen(all_args, **kwargs)
        stdout, stderr = p.communicate()
        self.log.debug(stdout)
        stdout = stdout.strip()
        if stderr is not None:
            stderr = stderr.strip()
            self.log.debug(stderr)
        exit_code = p.poll()
        result = ProcessResult(all_args, exit_code, stdout, stderr)
        self.log.debug("process: %s (%d)", result.cmd, result.exit_code)
        if self._exit_on_error:
            result.exit_on_error()
        return result


command = Process


class Commander(object):
    def __init__(self, log=log):
        self.log = log

    def set_log(self, logger):
        self.log = logger

    def __getattr__(self, key):
        return command((key,), log=self.log)

    def check(self, *args, **kwargs):
        kwargs.update({'log': self.log})
        return command(command=args, **kwargs).exit_on_error()

    def __call__(self, *args, **kwargs):
        kwargs.update({'log': self.log})
        return command(command=args, shell=True, **kwargs)


sh = Commander()
dig = Process(('dig', '+short'))
api_endpoints = Process(('juju', 'api-endpoints'))


def wait_for(timeout, interval, *callbacks, **kwargs):
    """
    Repeatedly try callbacks until all return True

    This will wait interval seconds between attempts and will error out
    after timeout has been exceeded.

    Callbacks will be called with the container as their argument.

    Setting timeout to zero will loop until cancelled, power runs outs,
    hardware fails, or the heat death of the universe.
    """
    start = time.time()
    if timeout:
        end = start + timeout
    else:
        end = 0

    bar = kwargs.get('bar', None)
    message = kwargs.get('message', None)
    once = 1
    while True:
        passes = True
        if end > 0 and time.time() > end:
            raise OSError("Timeout exceeded in wait_for")
        if bar:
            bar.next(once, message=message)
            if once == 1:
                once = 0
        if int(time.time()) % interval == 0:
            for callback in callbacks:
                result = callback()
                passes = passes & bool(result)
                if passes is False:
                    break
            if passes is True:
                break
        time.sleep(1)


def until(*callbacks, **kwargs):
    return wait_for(0, 20, *callbacks, **kwargs)


def retry(attempts, *callbacks, **kwargs):
    """
    Repeatedly try callbacks a fixed number of times or until all return True
    """
    for attempt in range(attempts):
        if 'bar' in kwargs:
            kwargs['bar'].next(attempt == 0, message=kwargs.get('message'))
        for callback in callbacks:
            if not callback():
                break
        else:
            break
    else:
        raise OSError("Retry attempts exceeded")
    return True


def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for fpath in os.environ["PATH"].split(os.pathsep):
            fpath = fpath.strip('"')
            exe_file = os.path.join(fpath, program)
            if is_exe(exe_file):
                return exe_file
    return None


def load_class(full_class_string, workingdir=None):
    """
    dynamically load a class from a string
    """

    module_path, class_str = full_class_string.rsplit(".", 1)
    if not workingdir:
        workingdir = os.getcwd()
    sys.path.insert(0, workingdir)
    try:
        module = importlib.import_module(module_path)
        return getattr(module, class_str)
    finally:
        sys.path.pop(0)


def walk(pathobj, fn, matcher=None, kind=None, **kwargs):
    """walk pathobj calling fn on each matched entry yielding each
    result. If kind is 'file' or 'dir' only that type ofd entry will
    be walked. matcher is an optional function returning bool indicating
    if the entry should be processed.
    """
    p = path(pathobj)
    walker = p.walk
    if kind == "files":
        walker = p.walkfiles
    elif kind == "dir":
        walker = p.walkdir

    for entry in walker():
        relpath = entry.relpath(pathobj)
        if matcher and not matcher(relpath):
            continue
        yield (entry, fn(entry, **kwargs))


def ignore_matcher(ignores=[]):
    spec = pathspec.PathSpec.from_lines(pathspec.GitIgnorePattern, ignores)

    def matcher(entity):
        return entity not in spec.match_files((entity,))
    return matcher


def sign(pathobj):
    p = path(pathobj)
    if not p.isfile():
        return None
    return hashlib.sha256(p.bytes()).hexdigest()


def delta_signatures(manifest_filename, ignore=None):
    md = path(manifest_filename)
    repo = md.normpath().dirname()

    expected = json.load(md.open())
    current = {}
    for rel, sig in walk(repo, sign, ignore_matcher(ignore or [])):
        rel = rel.relpath(repo)
        current[rel] = sig
    add, change, delete = set(), set(), set()

    for p, s in current.items():
        fp = repo / p
        if not fp.isfile():
            continue

        if p not in expected["signatures"]:
            add.add(p)
            continue
        # layer, kind, sig
        # don't include items generated only for the last layer
        if expected["signatures"][p][0] == "build":
            continue
        if expected["signatures"][p][2] != s:
            change.add(p)

    for p, d in expected["signatures"].items():
        if p not in current:
            delete.add(path(p))
    return add, change, delete


class ColoredFormatter(logging.Formatter):

    def __init__(self, terminal, *args, **kwargs):
        super(ColoredFormatter, self).__init__(*args, **kwargs)
        self._terminal = terminal

    def format(self, record):
        output = super(ColoredFormatter, self).format(record)
        if record.levelno >= logging.CRITICAL:
            line_color = self._terminal.bold_yellow_on_red
        elif record.levelno >= logging.ERROR:
            line_color = self._terminal.red
        elif record.levelno >= logging.WARNING:
            line_color = self._terminal.yellow
        elif record.levelno >= logging.INFO:
            line_color = self._terminal.green
        else:
            line_color = self._terminal.cyan
        return line_color(output)


class TermWriter(object):
    def __init__(self, fp=None, term=None, force_styling=False):
        if fp is None:
            fp = sys.stdout
        self.fp = fp
        if term is None:
            term = blessings.Terminal(force_styling=force_styling)
        self.term = term

    def __getattr__(self, key):
        return getattr(self.term, key)

    def write(self, msg, *args, **kwargs):
        if 't' in kwargs:
            raise ValueError("Using reserved token 't' in TermWriter.write")
        kwargs['t'] = self.term
        self.fp.write(msg.format(*args, **kwargs))


class _O(dict):
    def __getattr__(self, k):
        return self[k]


REACTIVE_PATTERNS = [
    re.compile(r"\s*@when"),
    re.compile(r".set_state\(")
]


def delta_python(orig, dest, patterns=REACTIVE_PATTERNS, context=2):
    """Delta two python files looking for certain patterns"""
    try:
        if isinstance(orig, path):
            od = orig.text()
        elif hasattr(orig, 'read'):
            od = orig.read()
        else:
            raise TypeError("Expected path() or file(), got %s" % type(orig))
    except Exception:
        log.error('Error reading {}'.format(orig))
        raise
    try:
        if isinstance(dest, path):
            dd = dest.text()
        elif hasattr(orig, 'read'):
            dd = dest.read()
        else:
            raise TypeError("Expected path() or file(), got %s" % type(dest))
    except Exception:
        log.error('Error reading {}'.format(dest))
        raise

    differ = diff_match_patch()
    linect = 0
    lastMatch = None
    for res in differ.diff_main(od, dd):
        if res[0] == diff_match_patch.DIFF_EQUAL:
            linect += res[1].count('\n')
            lastMatch = res[:]
            continue
        elif res[0] == diff_match_patch.DIFF_INSERT:
            linect += res[1].count('\n')
        else:
            linect -= res[1].count('\n')

        for p in patterns:
            if p.search(lastMatch[1]):
                yield [linect, lastMatch, res]
                break


def delta_python_dump(orig, dest, patterns=REACTIVE_PATTERNS,
                      context=2, term=None,
                      from_name=None, to_name=None):
    if term is None:
        term = TermWriter()

    def norm_sources(orig, dest):
        if from_name:
            oname = from_name
        else:
            oname = orig
        if to_name:
            dname = to_name
        else:
            dname = dest
        return _O({'orig_name': oname, 'dest_name': dname})

    def prefix_lines(lines, lineno):
        if isinstance(lines, str):
            lines = lines.splitlines()
        for i, l in enumerate(lines):
            lines[i] = "%-5d| %s" % (lineno + i, l)
        return "\n".join(lines)

    i = 0
    for lineno, last, current in delta_python(orig, dest, patterns, context):
        # pull enough context
        if last:
            context_lines = last[1].splitlines()[-context:]
        message = norm_sources(orig, dest)
        message['context'] = prefix_lines(context_lines, lineno - context)
        message['lineno'] = lineno
        message['delta'] = current[1]
        s = {diff_match_patch.DIFF_EQUAL: term.normal,
             diff_match_patch.DIFF_INSERT: term.green,
             diff_match_patch.DIFF_DELETE: term.red}[current[0]]
        message['status_color'] = s
        # output message
        term.write("{t.bold}{m.orig_name}{t.normal} --> "
                   "{t.bold}{m.dest_name}{t.normal}:\n",
                   m=message)
        term.write("{m.context}{m.status_color}{m.delta}{t.normal}\n",
                   m=message)
        i += 1
    return i == 0


class Description(argparse._StoreTrueAction):
    """A argparse action that prints parent parser's description & exits."""
    def __call__(self, parser, namespace, values, option_string=None):
        print(parser.description.split('\n')[0].strip('. '))
        raise SystemExit()


def add_plugin_description(parser):
    parser.add_argument('--description', action=Description)


class OrderedSet(collections.MutableSet):
    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]
        self.map = {}
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)


def validate_display_name(entity, linter):
    """Validate the display name info in entity metadata.

    :param entity: dict of entity metadata parsed from metadata.yaml
    :param linter: :class:`CharmLinter` or :class: `BundleLinter` object to
        which info/warning/error messages will be written
    """
    if 'display-name' not in entity:
        linter.info('`display-name` not provided, '
                    'add for custom naming in the UI')
        return

    # Keep display name in sync with the juju/names package.
    # https://github.com/juju/names/blob/master/charm.go#L33
    match = re.match(
        r'^[A-Za-z0-9]+( [-A-Za-z0-9_]|[-A-Za-z0-9_])*$',
        entity['display-name'])
    if not match:
        linter.err('display-name: not in valid format. '
                   'Only letters, numbers, dashes, and hyphens are permitted.')
        return



from __future__ import print_function
from __future__ import absolute_import

import os
import sys
import json
import argparse
from pkg_resources import resource_string, resource_exists

from charmtools.cli import parser_defaults
from charmtools import utils
from vergit import format_version


def get_args(args=None):
    parser = argparse.ArgumentParser(
        description='display tooling version information')
    parser.add_argument('--format',
                        choices=['long', 'short', 'default', 'json'],
                        default='default',
                        help="Version format. Long includes git revision "
                             "info. Default uses long if it's a pre-release.")
    utils.add_plugin_description(parser)
    parser = parser_defaults(parser)
    args = parser.parse_args(args)

    return args


def _add_snap_rev(version_info):
    if 'SNAP_REVISION' in os.environ:
        version_info['snap'] = '+snap-{}'.format(os.environ['SNAP_REVISION'])
    return version_info


def cached_charmstore_client_version():
    if 'SNAP' in os.environ:
        cscv = os.path.join(os.environ['SNAP'], 'charmstore-client-version')
        if not os.path.exists(cscv):
            return {'version': 'unavailable', 'git': ''}
        with open(cscv) as f:
            res_string = f.read().strip()
        return _add_snap_rev(json.loads(res_string))
    try:
        from apt.cache import Cache
        charm_vers = Cache()['charm'].versions
        for v in charm_vers:
            if v.is_installed:
                charm_ver = v.version
                break
        else:
            charm_ver = 'unavailable'
    except ImportError:
        charm_ver = 'unavailable'
    except:
        charm_ver = 'error'

    return _add_snap_rev({'version': charm_ver})


def cached_charm_tools_version():
    ctv = os.path.join(os.environ.get('SNAP', ''), 'charm-tools-version')
    if os.path.exists(ctv):
        with open(ctv) as f:
            res_string = f.read().strip()
        return _add_snap_rev(json.loads(res_string))
    if resource_exists(__name__, 'VERSION'):
        res_string = resource_string(__name__, 'VERSION')
        if sys.version_info >= (3, 0):
            res_string = res_string.decode('UTF-8')
        return _add_snap_rev(json.loads(res_string))
    if os.environ.get('SNAPCRAFT_PROJECT_VERSION', 'git') != 'git':
        version_parts = os.environ['SNAPCRAFT_PROJECT_VERSION'].split('+')
        git = ''
        gitn = 0
        if len(version_parts) > 1:
            git = version_parts[1]
            gitn = int(git.split('-')[1])
        return _add_snap_rev({
            'version': version_parts[0],
            'git': '+{}'.format(git),
            'gitn': gitn,
        })
    return {'version': 'unavailable'}


def main():
    args = get_args()

    if args.format == 'json':
        print(json.dumps({
            'charmstore-client': cached_charmstore_client_version(),
            'charm-tools': cached_charm_tools_version(),
        }))
    else:
        print("charmstore-client {}".format(
            format_version(cached_charmstore_client_version(), args.format)))
        print("charm-tools {}".format(
            format_version(cached_charm_tools_version(), args.format)))


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import sys
import mock
import unittest
from pkg_resources import resource_filename

# allow importing actions from the hooks directory
sys.path.append(resource_filename(__name__, '../hooks'))
import actions


class TestActions(unittest.TestCase):
    @mock.patch('charmhelpers.core.hookenv.log')
    def test_log_start(self, log):
        actions.log_start('test-service')
        log.assert_called_once_with('$metadata.package starting')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
#
#    Copyright (C) 2014  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import os.path as path
import time
import sys
import shutil
import tempfile

from Cheetah.Template import Template
from stat import ST_MODE

from charmtools.generators import (
    CharmTemplate,
)

log = logging.getLogger(__name__)


class PythonServicesCharmTemplate(CharmTemplate):
    """Creates a python-based charm using the services framework"""

    def create_charm(self, config, output_dir):
        self._copy_files(output_dir)

        for root, dirs, files in os.walk(output_dir):
            for outfile in files:
                if self.skip_template(outfile):
                    continue

                self._template_file(config, path.join(root, outfile))

    def _copy_files(self, output_dir):
        here = path.abspath(path.dirname(__file__))
        template_dir = path.join(here, 'files')
        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)
        shutil.copytree(template_dir, output_dir)

    def _template_file(self, config, outfile):
        if path.islink(outfile):
            return

        mode = os.stat(outfile)[ST_MODE]
        t = Template(file=outfile, searchList=(config))
        o = tempfile.NamedTemporaryFile(
            dir=path.dirname(outfile), delete=False)
        os.chmod(o.name, mode)
        st = str(t)
        if sys.version_info >= (3, ):
            st = st.encode('UTF-8')
        o.write(st)
        o.close()
        backupname = outfile + str(time.time())
        os.rename(outfile, backupname)
        os.rename(o.name, outfile)
        os.unlink(backupname)



#!/usr/bin/python

#    Copyright (C) 2019  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import argparse
import json
import os
import sys
from subprocess import run, PIPE


def _fail(msg, debug_msg=None, show_debug=False):
    if show_debug:
        if isinstance(debug_msg, bytes):
            debug_msg = debug_msg.decode('utf8')
        print(debug_msg, file=sys.stderr)
    print(msg, file=sys.stderr)
    sys.exit(1)


def _bin(filename):
    return os.path.join(os.environ['SNAP'], 'bin', filename)


def _url(charm_id):
    charm_id = charm_id[len('cs:'):]
    return f"https://api.jujucharms.com/charmstore/v5/{charm_id}/promulgate"


def promulgate():
    parser = argparse.ArgumentParser(
        description='Promote a charm from a personal namespace to the flat, '
                    'top-level namespace')
    parser.add_argument('charm_id',
                        help='Charm store URL of charm (e.g., cs:~user/charm)')
    parser.add_argument('--debug', action='store_true',
                        help='Output debug information')
    args = parser.parse_args()
    charm_id = args.charm_id

    if not charm_id.startswith('cs:'):
        _fail('Must provide a fully-qualified charm ID starting with "cs:"')
    if not charm_id.startswith('cs:~'):
        _fail('Must provide a namespaced charm ID')

    charm_show = run([_bin('wrappers/charm'), 'show',
                      charm_id, '--format=json'],
                     stdout=PIPE, stderr=PIPE)
    if charm_show.returncode != 0:
        _fail(f'{charm_id} is not released to stable, or cannot be read',
              charm_show.stdout + charm_show.stderr, args.debug)
    charm_info = json.loads(charm_show.stdout.decode('utf8'))
    if charm_info['promulgated']['Promulgated']:
        _fail(f'{charm_id} is already promulgated',
              charm_info['promulgated'], args.debug)
    if 'everyone' not in charm_info['perm']['Read']:
        _fail(f'{charm_id} is not readable by "everyone"',
              charm_info['perm']['Read'], args.debug)

    promulgated_id = 'cs:' + charm_id.split('/', 1)[-1]
    promulgated_show = run([_bin('wrappers/charm'), 'show',
                            promulgated_id, '--format=json'],
                           stdout=PIPE, stderr=PIPE)
    if promulgated_show.returncode == 0:
        promulgated_info = json.loads(promulgated_show.stdout.decode('utf8'))
        owner = promulgated_info['owner']['User']
        _fail(f'{promulgated_id} is already owned by {owner}; '
              f'you will need to unpromulgate first to change owner')

    res = run([_bin('bhttp'), 'put', '-j', _url(charm_id),
               'Promulgated:=true'])
    if res.returncode != 0:
        sys.exit(res.returncode)
    print(f'{charm_id} promulgated to {promulgated_id}')

    writers = [writer
               for writer in charm_info['perm']['Write']
               if writer not in ('charmers', 'everyone')]
    if not writers:
        _fail('Unable to determine non-charmers writer to grant; '
              'check and correct perms manually')
    for writer in writers:
        res = run([_bin('wrappers/charm'), 'grant',
                   promulgated_id, '--acl=write', writer],
                  stdout=PIPE, stderr=PIPE)
        if res.returncode != 0:
            _fail(f'Failed to grant write on {promulgated_id} to {writer}',
                  res.stdout + res.stderr, args.debug)
    print(f'Granted write on {promulgated_id} to: {", ".join(writers)}')


def unpromulgate():
    parser = argparse.ArgumentParser(
        description='Remove a charm from the flat, top-level namespace')
    parser.add_argument('charm_id',
                        help='Charm store URL of charm (e.g., cs:~user/charm)')
    parser.add_argument('--debug', action='store_true',
                        help='Output debug information')
    args = parser.parse_args()
    charm_id = args.charm_id

    if not charm_id.startswith('cs:'):
        _fail('Must provide a fully-qualified charm ID starting with "cs:"')
    if charm_id.startswith('cs:~'):
        owner, charm_id = charm_id[len('cs:~'):].split('/', 1)
        charm_id = 'cs:' + charm_id
    else:
        owner = None

    charm_show = run([_bin('wrappers/charm'), 'show',
                      charm_id, '--format=json'],
                     stdout=PIPE, stderr=PIPE)
    if charm_show.returncode != 0:
        _fail(f'{charm_id} is not promulgated',
              charm_show.stdout + charm_show.stderr, args.debug)

    charm_info = json.loads(charm_show.stdout.decode('utf8'))
    if owner is not None and charm_info['owner']['User'] != owner:
        _fail(f'{charm_id} is not owned by {owner}',
              charm_info['owner'], args.debug)

    res = run([_bin('bhttp'), 'put', '-j', _url(charm_id),
               'Promulgated:=false'])
    if res.returncode != 0:
        sys.exit(1)
    print(f'{charm_id} unpromulgated')



#!/usr/bin/env python
#
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).

import os
import subprocess
import sys
import json
from setuptools import setup, find_packages


curdir = os.path.dirname(__file__)
version_cache = os.path.join(curdir, 'charmtools', 'VERSION')
try:
    version_raw = subprocess.check_output(['vergit', '--format=json']).strip()
    if sys.version_info >= (3, 0):
        version_raw = version_raw.decode('UTF-8')
    version = json.loads(version_raw)['version']
except Exception:
    version = 'unknown'
if version == 'unknown':
    # during install; use cached VERSION
    try:
        with open(version_cache, 'r') as fh:
            version_raw = fh.read()
        version = json.loads(version_raw)['version']
    except Exception:
        version = None
else:
    # during build; update cached VERSION
    with open(version_cache, 'w') as fh:
        fh.write(version_raw)

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as fh:
    readme = fh.read()


setup(
    name='charm-tools',
    version=version,
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=[
        'launchpadlib<1.11',
        'cheetah3>=3.0.0',
        'pyyaml>=3.11,<4.3',
        'requests>=2.0.0,<3.0.0',
        'libcharmstore',
        'blessings<=1.6',
        'ruamel.yaml<0.16.0',
        'pathspec<=0.3.4',
        'otherstuf<=1.1.0',
        # 'path.py<=8.1.2',
        'path.py>=10.5',
        'pip>=1.5.4',
        # 'jujubundlelib',
        'virtualenv>=1.11.4',
        'colander<=1.7.0',
        'jsonschema<=2.5.1',
        'keyring<21',
        'secretstorage<2.4',
        'dict2colander==0.2',
        'vergit>=1.0.0,<2.0.0',
        'requirements-parser',
    ],
    include_package_data=True,
    maintainer='Cory Johns',
    maintainer_email='johnsca@gmail.com',
    description=('Tools for building and maintaining Juju charms'),
    long_description=readme,
    license='GPL v3',
    url='https://github.com/juju/charm-tools',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
    ],
    entry_points={
        'console_scripts': [
            'charm-add = charmtools.generate:main',
            'charm-build = charmtools.build.builder:main',
            'charm-create = charmtools.create:main',
            'charm-help = charmtools.cli:usage',
            'charm-layers = charmtools.build.builder:inspect',
            'charm-proof = charmtools.proof:main',
            'charm-pull-source = charmtools.pullsource:main',
            'charm-version = charmtools.version:main',
            'charm-promulgate = charmtools.promulgation:promulgate',
            'charm-unpromulgate = charmtools.promulgation:unpromulgate',
        ],
        'charmtools.templates': [
            'bash = charmtools.templates.bash:BashCharmTemplate',
            'reactive-python = charmtools.templates.reactive_python:ReactivePythonCharmTemplate',  # noqa: E501
            'reactive-bash = charmtools.templates.reactive_bash:ReactiveBashCharmTemplate',  # noqa: E501
            'python-basic = charmtools.templates.python:PythonCharmTemplate',
            'python = charmtools.templates.python_services:PythonServicesCharmTemplate',  # noqa: E501
            'ansible = charmtools.templates.ansible:AnsibleCharmTemplate',
            'chef = charmtools.templates.chef:ChefCharmTemplate',
            'powershell = charmtools.templates.powershell:PowerShellCharmTemplate',  # noqa: E501
        ]
    },
)



# coding=utf-8
import json
from ruamel import yaml
from charmtools.build import config
from charmtools import utils

theme = {
    -1: "bright_black",
    0: "normal",
    1: "green",
    2: "cyan",
    3: "magenta",
    4: "yellow",
    5: "red",
    6: "bold",
    7: "bright_magenta",
    8: "bright_red",
    9: "bright_cyan",
    10: "bright_green",
    11: "bright_yellow",
}


def scan_for(col, cur, depth):
    for e, (rel, d) in col[cur:]:
        if d and d == depth:
            return True
    return False


def get_prefix(walk, cur, depth, next_depth):
    guide = []
    for i in range(depth):
        # scan forward in walk from i seeing if a subsequent
        # entry happens at each depth
        if scan_for(walk, cur, i):
            guide.append(" │  ")
        else:
            guide.append("    ")
    if depth == next_depth:
        prefix = " ├─── "
    else:
        prefix = " └─── "
    return "{}{}".format("".join(guide), prefix)


def inspect(charm, force_styling=False, annotate=False):
    tw = utils.TermWriter(force_styling=force_styling)
    manp = charm / ".build.manifest"
    comp = charm / "layer.yaml"
    if not manp.exists() or not comp.exists():
        return
    manifest = json.loads(manp.text())
    composer = yaml.safe_load(comp.open())
    a, c, d = utils.delta_signatures(manp)

    # ordered list of layers used for legend
    if isinstance(manifest['layers'][0], dict):
        layers = [layer['url'] for layer in manifest['layers']]
    else:
        layers = list(manifest['layers'])
    layers.reverse()
    while layers[0].startswith('interface:'):
        layers.append(layers.pop(0))

    def get_depth(e):
        rel = e.relpath(charm)
        depth = len(rel.splitall()) - 2
        return rel, depth

    def get_suffix(rel):
        suffix = ""
        if rel in a:
            suffix = "+"
        elif rel in c:
            suffix = "*"
        return suffix

    def get_color(rel):
        # name of layer this belongs to
        color = tw.term.normal
        if rel in manifest['signatures']:
            layer = manifest['signatures'][rel][0]
            if layer in layers:
                layer_key = layers.index(layer)
            else:
                # handle special build created artifacts, which have
                # a "layer name" of "build" (mostly the manifest itself)
                layer_key = -1
            color = getattr(tw, theme.get(layer_key, "normal"))
        else:
            if entry.isdir():
                color = tw.blue
        return color

    tw.write("Inspect %s\n" % composer["is"])
    if tw.does_styling or force_styling:
        tw.write("\n")
        tw.write("Color key:\n")
        for i, layer in enumerate(layers):
            tw.write("# {color}{layer}{t.normal}\n",
                     color=getattr(tw, theme.get(i, "normal")),
                     layer=layer)
    else:
        # force annotations if we can't use color
        annotate = True
    tw.write("\n")
    tw.write("{t.blue}{target}{t.normal}\n", target=charm)

    ignorer = utils.ignore_matcher(config.DEFAULT_IGNORES)
    walk = sorted(utils.walk(charm, get_depth),
                  key=lambda x: x[1][0])
    for i in range(len(walk) - 1):
        entry, (rel, depth) = walk[i]
        nEnt, (nrel, ndepth) = walk[i + 1]
        if not ignorer(rel):
            continue

        if annotate and rel in manifest['signatures']:
            layer_name = manifest['signatures'][rel][0]
            if layer_name == 'build':
                # handle special build created artifacts, which have
                # a "layer name" of "build" (mostly the manifest itself)
                annotation = ' ({}build artifact{})'.format(tw.bright_black,
                                                            tw.normal)
            else:
                annotation = ' (from {}{}{})'.format(get_color(rel),
                                                     layer_name,
                                                     tw.normal)
        else:
            annotation = ''
        tw.write("{prefix}{layerColor}{entry} "
                 "{t.bold}{suffix}{t.normal}{annotation}\n",
                 prefix=get_prefix(walk, i, depth, ndepth),
                 layerColor=get_color(rel),
                 suffix=get_suffix(rel),
                 entry=rel.name,
                 annotation=annotation)
